# encoding: utf-8
"""
leuvenmapmatching.map.base
~~~~~~~~~~~~~~~~~~~~~~~~~~

Base Map class.

To be used in a Matcher object, the following functions need to be defined:

- ``edges_closeto``
- ``nodes_closeto``
- ``nodes_nbrto``
- ``edges_nbrto``

For visualiation purposes the following methods need to be implemented:

- ``bb``
- ``labels``
- ``size``
- ``coordinates``
- ``node_coordinates``
- ``all_edges``
- ``all_nodes``

:author: Wannes Meert
:copyright: Copyright 2018 DTAI, KU Leuven and Sirris.
:license: Apache License, Version 2.0, see LICENSE for details.
"""

from abc import abstractmethod
import logging
MYPY = False
if MYPY:
    from typing import Tuple, Union, List
    LabelType = Union[int, str]
    LocType = Tuple[float, float]


logger = logging.getLogger("be.kuleuven.cs.dtai.mapmatching")


class BaseMap(object):
    """Abstract class for a Map."""

    def __init__(self, name, use_latlon=True):
        """Simple database wrapper/stub."""
        self.name = name
        self._use_latlon = None
        self.distance = None
        self.distance_point_to_segment = None
        self.distance_segment_to_segment = None
        self.box_around_point = None
        self.use_latlon = use_latlon

    @property
    def use_latlon(self):
        return self._use_latlon

    @use_latlon.setter
    def use_latlon(self, value):
        self._use_latlon = value
        if self._use_latlon:
            from ..util import dist_latlon as dist_lib
        else:
            from ..util import dist_euclidean as dist_lib
        self.distance = dist_lib.distance
        self.distance_point_to_segment = dist_lib.distance_point_to_segment
        self.distance_segment_to_segment = dist_lib.distance_segment_to_segment
        self.box_around_point = dist_lib.box_around_point
        self.lines_parallel = dist_lib.lines_parallel

    @abstractmethod
    def bb(self):
        """Bounding box.

        :return: (lat_min, lon_min, lat_max, lon_max)
        """

    @abstractmethod
    def labels(self):
        """Labels of all nodes."""

    @abstractmethod
    def size(self):
        """Number of nodes."""

    @abstractmethod
    def node_coordinates(self, node_key):
        """Coordinates for given node key."""

    @abstractmethod
    def edges_closeto(self, loc, max_dist=None, max_elmt=None):
        """Find edges close to a certain location.

        :param loc: Latitude, Longitude
        :param max_dist: Maximal distance that returned nodes can be from lat-lon
        :param max_elmt: Maximal number of elements returned after sorting according to distance.
        :return: list[tuple[dist, label, loc]]
        """
        return None

    @abstractmethod
    def nodes_closeto(self, loc, max_dist=None, max_elmt=None):
        """Find nodes close to a certain location.

        :param loc: Latitude, Longitude
        :param max_dist: Maximal distance that returned nodes can be from lat-lon
        :param max_elmt: Maximal number of elements returned after sorting according to distance.
        :return: list[tuple[dist, label, loc]]
        """
        return None

    @abstractmethod
    def nodes_nbrto(self, node):
        # type: (BaseMap, LabelType) -> List[Tuple[LabelType, LocType]]
        """Return all nodes that are linked to ``node``.

        :param node: Node identifier
        :return: list[tuple[label, loc]]
        """
        return []

    def edges_nbrto(self, edge):
        # type: (BaseMap, Tuple[LabelType, LabelType]) -> List[Tuple[LabelType, LocType, LabelType, LocType]]
        """Return all edges that are linked to ``edge``.

        Defaults to ``nodes_nbrto``.

        :param edge: Edge identifier
        :return: list[tuple[label1, label2, loc1, loc2]]
        """
        results = []
        l1, l2 = edge
        p2 = self.node_coordinates(l2)
        for l3, p3 in self.nodes_nbrto(l2):
            results.append((l2, p2, l3, p3))
        return results

    @abstractmethod
    def all_nodes(self, bb=None):
        """All node keys and coordinates.

        :return: [(key, (lat, lon))]
        """

    @abstractmethod
    def all_edges(self, bb=None):
        """All edges.

        :return: [(key_a, loc_a, key_b, loc_b)]
        """



# encoding: utf-8
"""
leuvenmapmatching.util.segment
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:author: Wannes Meert
:copyright: Copyright 2015-2018 DTAI, KU Leuven and Sirris.
:license: Apache License, Version 2.0, see LICENSE for details.
"""
import logging


logger = logging.getLogger("be.kuleuven.cs.dtai.mapmatching")


class Segment(object):
    """Segment in the graph and its interpolated point."""
    __slots__ = ["l1", "p1", "l2", "p2", "_pi", "_ti"]

    def __init__(self, l1, p1, l2=None, p2=None, pi=None, ti=None):
        """Create a new segment.

        :param l1: Label of the node that is the start of the segment.
        :param p1: Point (coordinate) of the start node.
        :param l2: Label of the node that is the end of the segment.
        :param p2: Point (coordinate) of the end node.
        :param pi: Interpolated point. The point that is the best match and
            can be in between p1 and p2.
        :param ti: Position of pi on the segment [p1,p2], thus pi = p1+t1*(p2-p1).
        """
        self.l1 = l1  # Start of segment, label
        self.p1 = p1  # point
        self.l2 = l2  # End of segment, if None the segment is a point
        self.p2 = p2
        self.pi = pi  # Interpolated point
        self.ti = ti  # Position on segment p1-p2

    @property
    def label(self):
        if self.l2 is None:
            return self.l1
        return f"{self.l1}-{self.l2}"

    @property
    def rlabel(self):
        if self.l2 is None:
            return self.l1
        return f"{self.l2}-{self.l1}"

    @property
    def key(self):
        if self.l2 is None:
            return self.l1
        return f"{self.l1}-{self.l2}"

    @property
    def pi(self):
        if self.p2 is None:
            return self.p1
        return self._pi

    @pi.setter
    def pi(self, value):
        if value is not None and len(value) > 2:
            self._pi = tuple(value[:2])
        else:
            self._pi = value

    @property
    def ti(self):
        if self.p2 is None:
            return 0
        return self._ti

    @ti.setter
    def ti(self, value):
        self._ti = value

    def is_point(self):
        return self.p2 is None

    def last_point(self):
        if self.p2 is None:
            return self.p1
        return self.p2

    def loc_to_str(self):
        if self.p2 is None:
            return f"{self.p1}"
        if self._pi is not None:
            return f"{self.p1}-{self.pi}/{self.ti}-{self.p2}"
        return f"{self.p1}-{self.p2}"

    def __str__(self):
        if self.p2 is None:
            return f"{self.l1}"
        if self._pi is not None:
            return f"{self.l1}-i-{self.l2}"
        return f"{self.l1}-{self.l2}"

    def __repr__(self):
        return "Segment<" + self.__str__() + ">"



# encoding: utf-8
"""
leuvenmapmatching.util.dist_latlon
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Based on:
https://www.movable-type.co.uk/scripts/latlong.html
https://www.movable-type.co.uk/scripts/latlong-vectors.html

:author: Wannes Meert
:copyright: Copyright 2015-2018 DTAI, KU Leuven and Sirris.
:license: Apache License, Version 2.0, see LICENSE for details.
"""
import logging
import math
from math import radians, cos, sin, asin, acos, sqrt, atan2, fabs, degrees, ceil, copysign

from . import dist_euclidean as diste


logger = logging.getLogger("be.kuleuven.cs.dtai.mapmatching")
earth_radius = 6371000


def distance(p1, p2):
    """Distance between two points.

    :param p1: (Lat, Lon)
    :param p2: (Lat, Lon)
    :return: Distance in meters
    """
    lat1, lon1 = p1
    lat2, lon2 = p2
    lat1, lon1 = radians(lat1), radians(lon1)
    lat2, lon2 = radians(lat2), radians(lon2)
    dist = distance_haversine_radians(lat1, lon1, lat2, lon2)
    return dist


def distance_point_to_segment(p, s1, s2, delta=0.0, constrain=True):
    """Distance between point and segment.

    Cross-track distance.

    https://www.movable-type.co.uk/scripts/latlong.html#cross-track

    :param s1: Segment start point
    :param s2: Segment end point
    :param p: Point to measure distance from path to
    :param delta: Stay away from the endpoints with this factor
    :return: (Distance in meters, projected location on segment, relative location on segment)
    """
    lat1, lon1 = s1  # Start point
    lat2, lon2 = s2  # End point
    lat3, lon3 = p
    lat1, lon1 = radians(lat1), radians(lon1)
    lat2, lon2 = radians(lat2), radians(lon2)
    lat3, lon3 = radians(lat3), radians(lon3)

    dist_hs = distance_haversine_radians(lat1, lon1, lat2, lon2)
    if dist_hs == 0:
        dist_ct, pi, ti = distance(p, s1), s1, 0
        return dist_ct, pi, ti

    d13 = distance_haversine_radians(lat1, lon1, lat3, lon3)
    delta13 = d13 / earth_radius
    b13 = bearing_radians(lat1, lon1, lat3, lon3)
    b12 = bearing_radians(lat1, lon1, lat2, lon2)

    dxt = asin(sin(delta13) * sin(b13 - b12))
    # b13d12 = (b13 - b12) % (2 * math.pi)
    # if b13d12 > math.pi:
    #     b13d12 = 2 * math.pi - b13d12
    dist_ct = fabs(dxt) * earth_radius
    # Correct to negative value if point is before segment
    # sgn = -1 if b13d12 > (math.pi / 2) else 1
    sgn = copysign(1, cos(b12 - b13))
    dat = sgn * acos(cos(delta13) / abs(cos(dxt))) * earth_radius
    ti = dat / dist_hs

    if not constrain:
        lati, loni = destination_radians(lat1, lon1, b12, dat)
    elif ti > 1.0:
        ti = 1.0
        lati, loni = lat2, lon2
        dist_ct = distance_haversine_radians(lat3, lon3, lati, loni)
    elif ti < 0.0:
        ti = 0.0
        lati, loni = lat1, lon1
        dist_ct = distance_haversine_radians(lat3, lon3, lati, loni)
    else:
        lati, loni = destination_radians(lat1, lon1, b12, dat)
    pi = (degrees(lati), degrees(loni))

    return dist_ct, pi, ti


def distance_segment_to_segment(f1, f2, t1, t2):
    """Distance between segments. If no intersection within range, simplified to distance from f2 to [t1,t2].

    :param f1: From
    :param f2:
    :param t1: To
    :param t2:
    :return: (distance, proj on f, proj on t, rel pos on t)
    """
    # Translate lat-lon to x-y and apply the Euclidean function
    latf1, lonf1 = f1
    latf1, lonf1 = radians(latf1), radians(lonf1)
    f1 = 0, 0  # Origin

    latf2, lonf2 = f2
    latf2, lonf2 = radians(latf2), radians(lonf2)
    df1f2 = distance_haversine_radians(latf1, lonf1, latf2, lonf2)
    bf1f2 = bearing_radians(latf1, lonf1, latf2, lonf2)
    # print(f"bf1f2 = {bf1f2} = {degrees(bf1f2)} degrees")
    f2 = (df1f2 * cos(bf1f2),  df1f2 * sin(bf1f2))

    latt1, lont1 = t1
    latt1, lont1 = radians(latt1), radians(lont1)
    df1t1 = distance_haversine_radians(latf1, lonf1, latt1, lont1)
    bf1t1 = bearing_radians(latf1, lonf1, latt1, lont1)
    # print(f"bf1t1 = {bf1t1} = {degrees(bf1t1)} degrees")
    t1 = (df1t1 * cos(bf1t1), df1t1 * sin(bf1t1))

    latt2, lont2 = t2
    latt2, lont2 = radians(latt2), radians(lont2)
    dt1t2 = distance_haversine_radians(latt1, lont1, latt2, lont2)
    # print(f"dt1t2 = {dt1t2}")
    bt1t2 = bearing_radians(latt1, lont1, latt2, lont2)
    # print(f"bt1t2 = {bt1t2} = {degrees(bt1t2)} degrees")
    t2 = (t1[0] + dt1t2 * cos(bt1t2), t1[1] + dt1t2 * sin(bt1t2))

    d, pf, pt, u_f, u_t = diste.distance_segment_to_segment(f1, f2, t1, t2)
    pf = destination_radians(latf1, lonf1, bf1f2, u_f * df1f2)
    pf = (degrees(pf[0]), degrees(pf[1]))
    pt = destination_radians(latt1, lont1, bt1t2, u_t * dt1t2)
    pt = (degrees(pt[0]), degrees(pt[1]))

    return d, pf, pt, u_f, u_t


def project(s1, s2, p, delta=0.0):
    _, pi, ti = distance_point_to_segment(p, s1, s2, delta)
    return pi, ti


def box_around_point(p, dist):
    lat, lon = p
    latr, lonr = radians(lat), radians(lon)
    diag_dist = sqrt(2 * dist ** 2)
    lat_t, lon_r = destination_radians(latr, lonr, radians(45), diag_dist)
    lat_b, lon_l = destination_radians(latr, lonr, radians(225), diag_dist)
    lat_t, lon_r = degrees(lat_t), degrees(lon_r)
    lat_b, lon_l = degrees(lat_b), degrees(lon_l)
    return lat_b, lon_l, lat_t, lon_r


def interpolate_path(path, dd):
    """
    :param path: (lat, lon)
    :param dd: Distance difference (meter)
    :return:
    """
    path_new = [path[0]]
    for p1, p2 in zip(path, path[1:]):
        lat1, lon1 = p1[0], p1[1]
        lat2, lon2 = p2[0], p2[1]
        lat1, lon1 = radians(lat1), radians(lon1)
        lat2, lon2 = radians(lat2), radians(lon2)
        dist = distance_haversine_radians(lat1, lon1, lat2, lon2)
        if dist > dd:
            dt = int(ceil(dist / dd))
            distd = dist/dt
            disti = 0
            brng = bearing_radians(lat1, lon1, lat2, lon2)
            for _ in range(dt):
                disti += distd
                lati, loni = destination_radians(lat1, lon1, brng, disti)
                path_new.append((degrees(lati), degrees(loni)))
        path_new.append(p2)
    return path_new


def bearing_radians(lat1, lon1, lat2, lon2):
    """Initial bearing"""
    dlon = lon2 - lon1
    y = sin(dlon) * cos(lat2)
    x = cos(lat1) * sin(lat2) - sin(lat1) * cos(lat2) * cos(dlon)
    return atan2(y, x)


def distance_haversine_radians(lat1, lon1, lat2, lon2, radius=earth_radius):
    # type: (float, float, float, float, float) -> float
    lat = lat2 - lat1
    lon = lon2 - lon1
    a = sin(lat / 2) ** 2 + cos(lat1) * cos(lat2) * sin(lon / 2) ** 2
    # dist = 2 * radius * asin(sqrt(a))
    dist = 2 * radius * atan2(sqrt(a), sqrt(1 - a))
    return dist


def destination_radians(lat1, lon1, bearing, dist):
    d = dist / earth_radius
    lat2 = asin(sin(lat1) * cos(d) + cos(lat1) * sin(d) * cos(bearing))
    lon2 = lon1 + atan2(sin(bearing) * sin(d) * cos(lat1), cos(d) - sin(lat1) * sin(lat2))
    return lat2, lon2


def lines_parallel(f1, f2, t1, t2, d=None):
    latf1, lonf1 = f1
    latf1, lonf1 = radians(latf1), radians(lonf1)
    f1 = 0, 0  # Origin

    latf2, lonf2 = f2
    latf2, lonf2 = radians(latf2), radians(lonf2)
    df1f2 = distance_haversine_radians(latf1, lonf1, latf2, lonf2)
    bf1f2 = bearing_radians(latf1, lonf1, latf2, lonf2)
    # print(f"bf1f2 = {bf1f2} = {degrees(bf1f2)} degrees")
    f2 = (df1f2 * cos(bf1f2), df1f2 * sin(bf1f2))

    latt1, lont1 = t1
    latt1, lont1 = radians(latt1), radians(lont1)
    df1t1 = distance_haversine_radians(latf1, lonf1, latt1, lont1)
    bf1t1 = bearing_radians(latf1, lonf1, latt1, lont1)
    # print(f"bf1t1 = {bf1t1} = {degrees(bf1t1)} degrees")
    t1 = (df1t1 * cos(bf1t1), df1t1 * sin(bf1t1))

    latt2, lont2 = t2
    latt2, lont2 = radians(latt2), radians(lont2)
    dt1t2 = distance_haversine_radians(latt1, lont1, latt2, lont2)
    # print(f"dt1t2 = {dt1t2}")
    bt1t2 = bearing_radians(latt1, lont1, latt2, lont2)
    # print(f"bt1t2 = {bt1t2} = {degrees(bt1t2)} degrees")
    t2 = (t1[0] + dt1t2 * cos(bt1t2), t1[1] + dt1t2 * sin(bt1t2))

    return diste.lines_parallel(f1, f2, t1, t2, d=d)



# encoding: utf-8
"""
leuvenmapmatching.matcher.distance
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:author: Wannes Meert
:copyright: Copyright 2018 DTAI, KU Leuven and Sirris.
:license: Apache License, Version 2.0, see LICENSE for details.
"""
import logging
import math

from .base import BaseMatching, BaseMatcher
from ..util.segment import Segment
from ..util.debug import printd

MYPY = False
if MYPY:
    from typing import Tuple, Any, Dict


logger = logging.getLogger("be.kuleuven.cs.dtai.mapmatching")


class DistanceMatching(BaseMatching):
    __slots__ = ['d_s', 'd_o', 'lpe', 'lpt']  # Additional fields

    def __init__(self, *args, d_s=0.0, d_o=0.0, lpe=0.0, lpt=0.0, **kwargs):
        """

        :param args: Arguments for BaseMatching
        :param d_s: Distance between two (interpolated) states
        :param d_o: Distance between two (interpolated) observations
        :param lpe: Log probability of emission
        :param lpt: Log probablity of transition
        :param kwargs: Arguments for BaseMatching
        """
        super().__init__(*args, **kwargs)
        self.d_o: float = d_o
        self.d_s: float = d_s
        self.lpe: float = lpe
        self.lpt: float = lpt

    def _update_inner(self, m_other):
        # type: (DistanceMatching, DistanceMatching) -> None
        super()._update_inner(m_other)
        self.d_s = m_other.d_s
        self.d_o = m_other.d_o
        self.lpe = m_other.lpe
        self.lpt = m_other.lpt

    @staticmethod
    def repr_header(label_width=None, stop=""):
        res = BaseMatching.repr_header(label_width)
        res += f" {'dt(o)':<6} | {'dt(s)':<6} |"
        if logger.isEnabledFor(logging.DEBUG):
            res += f" {'lg(Pr-t)':<9} | {'lg(Pr-e)':<9} |"
        return res

    def __str__(self, label_width=None):
        res = super().__str__(label_width)
        res += f" {self.d_o:>6.2f} | {self.d_s:>6.2f} |"
        if logger.isEnabledFor(logging.DEBUG):
            res += f" {self.lpt:>9.2f} | {self.lpe:>9.2f} |"
        return res

    def __repr__(self):
        return self.label


class DistanceMatcher(BaseMatcher):
    """
    Map Matching that takes into account the distance between matched locations on the map compared to
    the distance between the observations (that are matched to these locations). It thus prefers matched
    paths that have a similar distance than the observations.

    Inspired on the method presented in:

        P. Newson and J. Krumm. Hidden markov map matching through noise and sparseness.
        In Proceedings of the 17th ACM SIGSPATIAL international conference on advances
        in geographic information systems, pages 336–343. ACM, 2009.

    The options available in :class:``BaseMatcher`` are inherited. Additionally, this class
    offers:

    - Transition probability is lower if the distance between observations and states is different
    - Transition probability is lower if the the next match is going back on an edge or to a previous edge
    - Transition probability is lower if two neighboring states represent not connected edges
    - Skip non-emitting states if distance between states and observations is close to each other
    """

    def __init__(self, *args, **kwargs):
        """Create a new object.

        :param map_con: Map object to connect to map database
        :param obs_noise: Standard deviation of noise
        :param obs_noise_ne: Standard deviation of noise for non-emitting states (is set to obs_noise if not given)
        :param max_dist_init: Maximum distance from start location (if not given, uses max_dist)
        :param max_dist: Maximum distance from path (this is a hard cut, min_prob_norm should be better)
        :param min_prob_norm: Minimum normalized probability of observations (ema)
        :param non_emitting_states: Allow non-emitting states. A non-emitting state is a state that is
            not associated with an observation. Here we assume it can be associated with a location in between
            two observations to allow for pruning. It is advised to set min_prob_norm and/or max_dist to avoid
            visiting all possible nodes in the graph.
        :param non_emitting_length_factor: Reduce the probability of a sequence of non-emitting states the longer it
            is. This can be used to prefer shorter paths. This is separate from the transition probabilities because
            transition probabilities are averaged for non-emitting states and thus the length is also averaged out.
        :param max_lattice_width: Restrict the lattice (or possible candidate states per observation) to this value.
            If there are more possible next states, the states with the best likelihood so far are selected.

        :param dist_noise: Standard deviation of difference between distance between states and distance
            between observatoins. If not given, set to obs_noise
        :param dist_noise_ne: If not given, set to dist_noise
        :param restrained_ne: Avoid non-emitting states if the distance between states and between
            observations is close to each other.
        :param avoid_goingback: If true, the probability is lowered for a transition that returns back to a
            previous edges or returns to a position on an edge.

        :param args: Arguments for BaseMatcher
        :param kwargs: Arguments for BaseMatcher
        """

        if not kwargs.get("only_edges", True):
            logger.warning("The MatcherDistance method only works on edges as states. Nodes have been disabled.")
        kwargs["only_edges"] = True
        if "matching" not in kwargs:
            kwargs["matching"] = DistanceMatching
        super().__init__(*args, **kwargs)
        self.use_original = kwargs.get('use_original', False)

        # if not use_original, the following value for beta gives a prob of 0.5 at dist=x_half:
        # beta = np.sqrt(np.power(x_half, 2) / (np.log(2)*2))
        self.dist_noise = kwargs.get('dist_noise', self.obs_noise)
        self.dist_noise_ne = kwargs.get('dist_noise_ne', self.dist_noise)
        self.beta = 2 * self.dist_noise ** 2
        self.beta_ne = 2 * self.dist_noise_ne ** 2

        self.sigma = 2 * self.obs_noise ** 2
        self.sigma_ne = 2 * self.obs_noise_ne ** 2

        self.restrained_ne = kwargs.get('restrained_ne', True)
        self.restrained_ne_thr = 1.25  # Threshold
        self.exact_dt_s = True  # Newson and Krumm is 'True'

        self.avoid_goingback = kwargs.get('avoid_goingback', True)
        self.gobackonedge_factor_log = math.log(0.5)
        self.gobacktoedge_factor_log = math.log(0.5)
        self.first_farend_penalty = math.log(0.75)  # should be > gobacktoedge_factor_log

        self.notconnectededges_factor_log = math.log(0.5)

    def logprob_trans(self, prev_m, edge_m, edge_o,
                      is_prev_ne=False, is_next_ne=False):
        # type: (DistanceMatcher, DistanceMatching, Segment, Segment, bool, bool) -> Tuple[float, Dict[str, Any]]
        """Transition probability.

        The probability is defined with a formula from the exponential family.
        :math:`P(dt) = exp(-d_t^2 / (2 * dist_{noise}^2))`

        with :math:`d_t = |d_s - d_o|,
        d_s = |loc_{prev\_state} - loc_{cur\_state}|,
        d_o = |loc_{prev\_obs} - loc_{cur\_obs}|`

        This function is more tolerant for low values. The intuition is that values under a certain
        distance should all be close to probability 1.0.

        Note: We should also smooth the distance between observations to handle outliers better.

        :param prev_m: Previous matching / state
        :param edge_m: Edge between matchings / states
        :param edge_o: Edge between observations
        :param is_prev_ne: Is previous state non-emitting
        :param is_next_ne: Is the next state non-emitting
        :param dist_o: First output of distance_progress
        :param dist_m: Second output of distance_progress
        :return:
        """
        d_z = self.map.distance(prev_m.edge_o.pi, edge_o.pi)
        is_same_edge = False
        if (prev_m.edge_m.l1 == edge_m.l1 and prev_m.edge_m.l2 == edge_m.l2) or \
                (prev_m.edge_m.l1 == edge_m.l2 and prev_m.edge_m.l2 == edge_m.l1):
            is_same_edge = True
        if ((not self.exact_dt_s) or
                is_same_edge or  # On same edge
                prev_m.edge_m.l2 != edge_m.l1):  # Edges are not connected
            d_x = self.map.distance(prev_m.edge_m.pi, edge_m.pi)
        else:
            # Take into account the curvature
            d_x = self.map.distance(prev_m.edge_m.pi, prev_m.edge_m.p2) + self.map.distance(prev_m.edge_m.p2, edge_m.pi)

        if is_next_ne:
            # For non-emitting states, the distances are added
            # Otherwise it can map to a sequence of short segments and stay at the same
            # observation because the difference is then always small.
            d_z += prev_m.d_o
            d_x += prev_m.d_s

        d_t = abs(d_z - d_x)
        # p_dt = 1 / beta * math.exp(-d_t / beta)
        if is_prev_ne or is_next_ne:
            beta = self.beta_ne
        else:
            beta = self.beta
        logprob = -d_t ** 2 / beta

        # Penalties
        if prev_m.edge_m.label == edge_m.label:
            # Staying in same state
            if self.avoid_goingback and edge_m.key == prev_m.edge_m.key and edge_m.ti < prev_m.edge_m.ti:
                # Going back on edge (direction is from p1 to p2 of the segment)
                logprob += self.gobackonedge_factor_log  # Prefer not going back
        elif (prev_m.edge_m.l1, prev_m.edge_m.l2) == (edge_m.l2, edge_m.l1):
            if self.avoid_goingback:
                logprob += self.gobackonedge_factor_log
        else:
            # Moving states
            if prev_m.edge_m.l2 != edge_m.l1:
                # We are moving between states that represent edges that are not connected through a node
                logprob += self.notconnectededges_factor_log
            elif self.avoid_goingback:
                # Goin back on state
                going_back = False
                for m in prev_m.prev:
                    if edge_m.label == m.edge_m.label:
                        going_back = True
                        break
                if going_back:
                    logprob += self.gobacktoedge_factor_log  # prefer not going back

        props = {
            'd_o': d_z,
            'd_s': d_x,
            'lpt': logprob
        }
        return logprob, props

    def logprob_obs(self, dist, prev_m=None, new_edge_m=None, new_edge_o=None, is_ne=False):
        # type: (DistanceMatcher, float, DistanceMatching, Segment, Segment, bool) -> Tuple[float, Dict[str, Any]]
        """Emission probability for emitting states.

        Exponential family:
        :math:`P(dt) = exp(-d_o^2 / (2 * obs_{noise}^2))`

        with :math:`d_o = |loc_{state} - loc_{obs}|`

        """
        if is_ne:
            sigma = self.sigma_ne
        else:
            sigma = self.sigma
        result = -dist ** 2 / sigma
        props = {
            'lpe': result
        }
        return result, props

    def _skip_ne_states(self, next_ne_m):
        # type: (DistanceMatcher, DistanceMatching) -> bool
        # Skip searching for non-emitting states when the distances between nodes
        # on the map are similar to the distances between the observation
        if not self.restrained_ne:
            return False
        if next_ne_m.d_s > 0:
            factor = (next_ne_m.d_o + next_ne_m.dist_obs) / next_ne_m.d_s
        else:
            factor = 0
        if factor < self.restrained_ne_thr:
            logger.debug(f"Skip non-emitting states to {next_ne_m.label}: {factor} < {self.restrained_ne_thr} "
                         "(observations close enough to each other)")
            return True
        return False



# encoding: utf-8
"""
leuvenmapmatching.util
~~~~~~~~~~~~~~~~~~~~~~

:author: Wannes Meert
:copyright: Copyright 2015-2018 DTAI, KU Leuven and Sirris.
:license: Apache License, Version 2.0, see LICENSE for details.
"""

# No automatic loading to avoid dependency on packages such as nvector and gpxpy if not used.


def approx_equal(a, b, rtol=0.0, atol=1e-08):
    return abs(a - b) <= (atol + rtol * abs(b))


def approx_leq(a, b, rtol=0.0, atol=1e-08):
    return (a - b) <= (atol + rtol * abs(b))



# encoding: utf-8
"""
leuvenmapmatching.map.sqlite
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Map representation based on a sqlite database. Not optimized for production purposes.

:author: Wannes Meert
:copyright: Copyright 2018 DTAI, KU Leuven and Sirris.
:license: Apache License, Version 2.0, see LICENSE for details.
"""
import sqlite3
import tempfile
import logging
import time
from pathlib import Path
import pickle
from functools import partial
try:
    import pyproj
except ImportError:
    pyproj = None
try:
    import tqdm
except ImportError:
    tqdm = None


from .base import BaseMap


MYPY = False
if MYPY:
    from typing import Optional, Set, Tuple, Dict, Union
    LabelType = Union[int, str]
    LocType = Tuple[float, float]
    EdgeType = Tuple[LabelType, LabelType]


logger = logging.getLogger("be.kuleuven.cs.dtai.mapmatching")


class SqliteMap(BaseMap):
    def __init__(self, name, use_latlon=True,
                 crs_lonlat=None, crs_xy=None, dir=None, deserializing=False):
        """Store a map as a SQLite instance.

        This class supports:

        - Indexing using rtrees to allow for fast searching of points on the map.
          When using the rtree index, only integer numbers are allowed as node labels.
        - Serializing to write and read from files.
        - Projecting points to a different frame (e.g. GPS to Lambert)

        :param name: Name of database file
        :param use_latlon: The locations represent latitude-longitude pairs, otherwise y-x coordinates
            are assumed.
        :param crs_lonlat: Coordinate reference system for the latitude-longitude coordinates.
        :param crs_xy: Coordiante reference system for the y-x coordinates.
        :param dir: Directory where to serialize to. If not given, a temporary location will be used.
        :param deserializing: Internal variable to indicate that the object is being build from a file.
        """
        super(SqliteMap, self).__init__(name, use_latlon=use_latlon)
        self.dir = Path(tempfile.gettempdir()) if dir is None else Path(dir)
        name = Path(name)
        suffix = name.suffix
        if suffix == '':
            name = name.with_suffix('.sqlite')
        self.db_fn = self.dir / name
        if deserializing and not self.db_fn.exists():
            raise Exception(f"File not found: {self.db_fn}")
        logger.debug(f"Opening database: {self.db_fn}")
        try:
            self.db = sqlite3.connect(str(self.db_fn))
        except Exception as exc:
            raise Exception(f'Problem with database: {self.db_fn}') from exc
        self.crs_lonlat = crs_lonlat
        self.crs_xy = crs_xy
        self.use_latlon = use_latlon

        if deserializing:
            self.read_properties()
        else:
            self.create_db()

        if self.crs_lonlat is None:
            self.crs_lonlat = 'EPSG:4326'  # GPS
        if self.crs_xy is None:
            self.crs_xy = 'EPSG:3395'  # Mercator projection

        self.save_properties()

        if pyproj:
            proj_lonlat = pyproj.Proj(self.crs_lonlat, preserve_units=True)
            proj_xy = pyproj.Proj(self.crs_xy, preserve_units=True)
            self.lonlat2xy = partial(pyproj.transform, proj_lonlat, proj_xy)
            self.xy2lonlat = partial(pyproj.transform, proj_xy, proj_lonlat)
        else:
            def pyproj_notfound(*_args, **_kwargs):
                raise Exception("pyproj package not found")
            self.lonlat2xy = pyproj_notfound
            self.xy2lonlat = pyproj_notfound

    def read_properties(self):
        c = self.db.cursor()
        for row in c.execute("SELECT key, value FROM properties;"):
            key, value = row[0], pickle.loads(row[1])
            self.__dict__[key] = value

    def save_properties(self):
        c = self.db.cursor()
        q = "INSERT INTO properties (key, value) VALUES (?, ?)"
        v = [('name', pickle.dumps(self.name)),
             ('use_latlon', pickle.dumps(self.use_latlon)),
             ('crs_lonlat', pickle.dumps(self.crs_lonlat)),
             ('crs_xy', pickle.dumps(self.crs_xy))]
        c.executemany(q, v)
        self.db.commit()

    def create_db(self):
        logger.debug("Cleaning database file and creating new tables")
        c = self.db.cursor()
        c.execute("DROP INDEX IF EXISTS edges_from_index")
        c.execute("DROP INDEX IF EXISTS close_edges_index")
        c.execute("DROP TABLE IF EXISTS nodes_index")
        c.execute("DROP TABLE IF EXISTS nodes")
        c.execute("DROP TABLE IF EXISTS edges_index")
        c.execute("DROP TABLE IF EXISTS edges")
        c.execute("DROP TABLE IF EXISTS close_edges")
        c.execute("DROP TABLE IF EXISTS properties")
        self.db.commit()

        # Create tables
        q = ("CREATE VIRTUAL TABLE nodes_index USING rtree(\n"
             "id,              -- Integer primary key\n"
             "minX, maxX,      -- Minimum and maximum X coordinate\n"
             "minY, maxY       -- Minimum and maximum Y coordinate\n"
             ")")
        c.execute(q)
        q = ("CREATE TABLE nodes(\n"
             "id INTEGER PRIMARY KEY,\n"
             "x REAL,\n"
             "y REAL\n"
             ")")
        c.execute(q)
        q = ("CREATE VIRTUAL TABLE edges_index USING rtree(\n"
             "id,              -- Integer primary key\n"
             "minX, maxX,      -- Minimum and maximum X coordinate\n"
             "minY, maxY       -- Minimum and maximum Y coordinate\n"
             ")")
        c.execute(q)
        q = ("CREATE TABLE edges(\n"
             "id INTEGER PRIMARY KEY,\n"
             "path INTEGER,\n"  # Not necessarily unique, a pathway id can consist of multiple edges
             "pathnum INTEGER,\n"
             "id1 INTEGER,\n"  # node 1
             "id2 INTEGER,\n"  # node 2
             "speed REAL,\n"  # speed m/s
             "type INTEGER\n"  # extra field
             ")")
        c.execute(q)
        q = ("CREATE TABLE close_edges(\n"
             "id1 INTEGER,\n"  # edge 1
             "id2 INTEGER\n"  # edge 2
             ")")
        c.execute(q)
        q = ("CREATE TABLE properties(\n"
             "key TEXT,\n"
             "value BLOB\n"
             ")")
        c.execute(q)
        q = "CREATE INDEX edges_from_index ON edges(id1)"
        c.execute(q)
        q = "CREATE INDEX close_edges_index ON close_edges(id1)"
        c.execute(q)
        self.db.commit()

    @classmethod
    def from_file(cls, filename):
        """Read from an existing file."""
        filename = Path(filename).with_suffix('')
        nmap = cls(filename.name, dir=filename.parent, deserializing=True)
        return nmap

    def bb(self):
        """Bounding box.

        :return: (lat_min, lon_min, lat_max, lon_max) or (y_min, x_min, y_max, x_max)
        """
        c = self.db.cursor()
        c.execute('SELECT min(minX), max(minX), min(maxX), max(maxX) FROM nodes_index;')
        lon_min, lon_max, lat_min, lat_max = c.fetchone()
        return lat_min, lon_min, lat_max, lon_max

    def labels(self):
        """All labels."""
        c = self.db.cursor()
        c.execute('SELECT id FROM nodes;')
        result = [row[0] for row in c.fetchall()]
        return result

    def size(self):
        c = self.db.cursor()
        c.execute('SELECT count(*) FROM nodes')
        result = c.fetchone()[0]
        return result

    def node_coordinates(self, node_key):
        """Get the coordinates of the given node.

        :param node_key: Node label/key
        :return: (lat, lon)
        """
        c = self.db.cursor()
        c.execute('SELECT y, x FROM nodes WHERE id = ?', (node_key, ))
        result = c.fetchone()
        if result is None:
            raise Exception(f"No coordinates found for node {node_key}")
        return result

    def add_node(self, node, loc, ignore_doubles=False, no_index=False, no_commit=False):
        """Add new node to the map.

        :param node: label
        :param loc: (lat, lon) or (y, x)
        :param ignore_doubles: When trying to add the same node, ignore it
        :param no_commit: Do not commit to database (remember to commit later)
        """
        c = self.db.cursor()
        lat, lon = loc
        # Nodes
        q = "INSERT INTO nodes VALUES(?, ?, ?)"
        try:
            c.execute(q, (node, lon, lat))
        except sqlite3.IntegrityError as exc:
            if ignore_doubles and "UNIQUE constraint failed: nodes.id" in str(exc):
                return
            logger.error(f"Problem with adding node {node} {loc}")
            raise exc
        # Nodes index
        if not no_index:
            q = "INSERT INTO nodes_index VALUES(?, ?, ?, ?, ?)"
            try:
                c.execute(q, (node, lon, lon, lat, lat))
            except sqlite3.IntegrityError as exc:
                logger.error(f"Problem with adding node to index {node} {loc}")
                raise exc
        if not no_commit:
            self.db.commit()

    def reindex_nodes(self):
        logger.debug("Reindexing nodes ...")
        t_start = time.time()
        c = self.db.cursor()
        c.execute('DELETE FROM nodes_index')
        q = ("INSERT INTO nodes_index "
             "SELECT id, x, x, y, y FROM nodes")
        c.execute(q)
        self.db.commit()
        c.execute('SELECT count(*) FROM nodes_index')
        cnt = c.fetchone()[0]
        t_delta = time.time() - t_start
        logger.debug(f"... done, #rows = {cnt}, time = {t_delta} sec")

    def add_nodes(self, nodes):
        """Add list of nodes to database.

        :param nodes: List[Tuple[node_key, Tuple[lat, lon]]]
        """
        c = self.db.cursor()

        def get_node_index():
            for key, (lat, lon) in nodes:
                yield key, lon, lon, lat, lat

        q = "INSERT INTO nodes_index VALUES(?, ?, ?, ?, ?)"
        c.executemany(q, get_node_index())

        def get_node_vals():
            for key, (lat, lon) in nodes:
                yield key, lon, lat

        q = "INSERT INTO nodes VALUES(?, ?, ?)"
        c.executemany(q, get_node_vals())
        self.db.commit()

    def del_node(self, node):
        raise Exception("TODO")

    def add_edge(self, node_a, node_b, loc_a=None, loc_b=None, speed=None, edge_type=None,
                 path=None, pathnum=None,
                 no_index=False, no_commit=False):
        """Add new edge to the map.

        :param node_a: Label for the node that is the start of the edge
        :param node_b: Label for the node that is the end of the edge
        :param no_commit: Do not commit to database (remember to commit later)
        """
        c = self.db.cursor()
        eid = (node_a, node_b).__hash__()
        c.execute('INSERT OR IGNORE INTO edges(id, path, pathnum, id1, id2, type, speed) VALUES (?, ?, ?, ?, ?, ?, ?)',
                  (eid, path, pathnum, node_a, node_b, edge_type, speed))
        # c.execute('SELECT last_insert_rowid();')
        # eid = c.fetchone()[0]

        if not no_index:
            if loc_a is None:
                c.execute('SELECT y, x FROM nodes WHERE id = ?;', (node_a, ))
                loc_a = c.fetchone()
            if loc_b is None:
                c.execute('SELECT y, x FROM nodes WHERE id = ?;', (node_b, ))
                loc_b = c.fetchone()
            lat1, lon1 = loc_a
            lat2, lon2 = loc_b
            if lat1 > lat2:
                lat1, lat2 = lat2, lat1
            if lon1 > lon2:
                lon1, lon2 = lon2, lon1
            c.execute('INSERT OR IGNORE INTO edges_index(id, minX, maxX, minY, maxY) VALUES (?, ?, ?, ?, ?)',
                      (eid, lon1, lon2, lat1, lat2))
        if not no_commit:
            self.db.commit()

    def add_edges(self, edges, no_index=False):
        """Add list of nodes to database.

        :param edges: List[Tuple[node_key, node_key]] or
            List[Tuple[node_key, node_key, path_key, int]]
        """
        c = self.db.cursor()

        def get_edge():
            for row in edges:
                row = list(row) + ([None] * (6 - len(row)))
                key_a, key_b, path, pathnum, edge_type, speed = row
                eid = (key_a, key_b).__hash__()
                yield eid, path, pathnum, key_a, key_b, edge_type, speed

        q = "INSERT INTO edges(id, path, pathnum, id1, id2, type, speed) VALUES(?, ?, ?, ?, ?, ?, ?);"
        c.executemany(q, get_edge())
        self.db.commit()

        if not no_index:
            self.reindex_edges()

    def reindex_edges(self):
        logger.debug("Reindexing edges ...")
        t_start = time.time()
        c = self.db.cursor()
        # c2 = self.db.cursor()
        c.execute('DELETE FROM edges_index')
        q = ('INSERT INTO edges_index '
             'SELECT e.id, MIN(n1.x,n2.x), MAX(n1.x,n2.x), '
             '             MIN(n1.y,n2.y), MAX(n1.y,n2.y) '
             'FROM edges e '
             'INNER JOIN nodes n1 ON n1.id = e.id1 '
             'INNER JOIN nodes n2 ON n2.id = e.id2')
        c.execute(q)
        # cnt = 0
        # for row in c.execute(q):
        #     # Contained in query
        #     c2.execute('INSERT INTO edges_index(id, minX, maxX, minY, maxY) VALUES (?, ?, ?, ?, ?)', row)
        #     cnt += 1
        self.db.commit()
        c.execute('SELECT count(*) FROM edges_index')
        cnt = c.fetchone()[0]
        t_delta = time.time() - t_start
        logger.debug(f"... done, #rows = {cnt}, time = {t_delta} sec")

    def all_edges(self, bb=None):
        """Return all edges.

        :param bb: Bounding box
        :return: (key_a, loc_a, nbr, loc_b)
        """
        c = self.db.cursor()
        q = 'SELECT e.id1, e.id2, n1.x AS n1x, n2.x AS n2x, n1.y AS n1y, n2.y AS n2y ' + \
            'FROM edges e, edges_index ei ' + \
            'LEFT JOIN nodes n1 ON n1.id = e.id1 ' + \
            'LEFT JOIN nodes n2 ON n2.id = e.id2 ' + \
            'WHERE ei.id == e.id'
        if bb:
            min_y, min_x, max_y, max_x = bb
            # Intersecting with query
            q += ' AND ei.maxX >= ? AND ei.minX <= ? AND ei.maxY >= ? AND ei.minY <= ?'
            c.execute(q, (min_x, max_x, min_y, max_y))
        else:
            c.execute(q)

        for row in c.fetchall():
            key_a, key_b, lon_a, lon_b, lat_a, lat_b = row
            yield key_a, (lat_a, lon_a), key_b, (lat_b, lon_b)

    def all_nodes(self, bb=None):
        """Return all nodes.

        :param bb: Bounding box (minY, minX, maxY, maxX)
        :return:
        """
        c = self.db.cursor()
        q = ('SELECT n.id, n.x, n.y '
             'FROM nodes n, nodes_index ni '
             'WHERE n.id = ni.id ')
        if bb:
            minY, minX, maxY, maxX = bb
            q += 'AND ni.minX >= ? AND ni.maxX <= ? AND ni.minY >= ? AND ni.maxY <= ?'
            c.execute(q, (minX, maxX, minY, maxY))
        else:
            c.execute(q)

        for row in c.fetchall():
            key_a, lon_a, lat_a = row
            yield key_a, (lat_a, lon_a)

    def purge(self):
        pass

    def to_xy(self, name=None):
        """Create a map that uses a projected XY representation on which Euclidean distances
        can be used.
        """
        if not self.use_latlon:
            return self
        if name is None:
            name = self.name + "_xy"

        logger.debug("Start transformation ...")
        t_start = time.time()
        nmap = self.__class__(name, dir=self.dir, use_latlon=self.use_latlon,
                              crs_xy=self.crs_xy, crs_lonlat=self.crs_lonlat)

        raise Exception("to implement")

        t_delta = time.time() - t_start
        logger.debug(f"... done: rtree size = {self.rtree_size()}, time = {t_delta} sec")

        return nmap

    def latlon2xy(self, lat, lon):
        x, y = self.lonlat2xy(lon, lat)
        return x, y

    def latlon2yx(self, lat, lon):
        x, y = self.lonlat2xy(lon, lat)
        return y, x

    def xy2latlon(self, x, y):
        lon, lat = self.xy2lonlat(x, y)
        return lat, lon

    def yx2latlon(self, y, x):
        lon, lat = self.xy2lonlat(x, y)
        return lat, lon

    def nodes_closeto(self, loc, max_dist=None, max_elmt=None):
        """Return all nodes close to the given location.

        :param loc: Location
        :param max_dist: Maximal distance from the location
        :param max_elmt: Return only the most nearby nodes
        """
        t_start = time.time()
        lat, lon = loc[:2]
        lat_b, lon_l, lat_t, lon_r = self.box_around_point((lat, lon), max_dist)
        bb = (lat_b, lon_l,  # y_min, x_min
              lat_t, lon_r)  # y_max, x_max
        nodes = self.all_nodes(bb=bb)
        t_delta_search = time.time() - t_start
        t_start = time.time()
        results = []
        for key_o, loc_o in nodes:
            dist = self.distance(loc, loc_o)
            if dist < max_dist:
                results.append((dist, key_o, loc_o))
        results.sort()
        t_delta_dist = time.time() - t_start
        logger.debug(f"Found {len(results)} closeby nodes "
                     f"in {t_delta_search} sec and computed distances in {t_delta_dist} sec")
        if max_elmt is not None:
            results = results[:max_elmt]
        return results

    def edges_closeto(self, loc, max_dist=None, max_elmt=None):
        """Return all nodes that are on an edge that is close to the given location.

        :param loc: Location
        :param max_dist: Maximal distance from the location
        :param max_elmt: Return only the most nearby nodes
        """
        print(f"edges_closeto({loc})")
        t_start = time.time()
        lat, lon = loc[:2]
        lat_b, lon_l, lat_t, lon_r = self.box_around_point((lat, lon), max_dist)
        bb = (lat_b, lon_l,  # y_min, x_min
              lat_t, lon_r)  # y_max, x_max
        logger.debug(f"Search in bounding box {bb}")
        nodes = self.all_edges(bb=bb)
        t_delta_search = time.time() - t_start
        t_start = time.time()
        results = []
        for key_a, loc_a, key_b, loc_b in nodes:
            dist, pi, ti = self.distance_point_to_segment(loc, loc_a, loc_b)
            if dist < max_dist:
                results.append((dist, key_a, loc_a, key_b, loc_b, pi, ti))
        results.sort()
        t_delta_dist = time.time() - t_start
        logger.debug(f"Found {len(results)} closeby edges "
                     f"in {t_delta_search} sec and computed distances in {t_delta_dist} sec")
        if max_elmt is not None:
            results = results[:max_elmt]
        return results

    def nodes_nbrto(self, node):
        c = self.db.cursor()
        q = ('SELECT e.id2, n2.y, n2.x FROM edges e '
             'INNER JOIN nodes n2 ON n2.id = e.id2 '
             'WHERE e.id1 = ?')
        results = []
        for nbr_label, nbr_lat, nbr_lon in c.execute(q, (node, )):
            results.append((nbr_label, (nbr_lat, nbr_lon)))
        return results

    def edges_nbrto(self, edge):
        l1, l2 = edge
        c = self.db.cursor()
        c.execute('SELECT n.y, n.x FROM nodes n WHERE id = ?', (l2, ))
        p2 = c.fetchone()
        results = []
        # Edges that connect at end of this edge
        for l3, p3 in self.nodes_nbrto(l2):
            results.append((l2, p2, l3, p3))
        # Edges that are in parallel and close
        edge_id = edge.__hash__()
        q = ('SELECT e.id1, e.id2, n1.y, n1.x, n2.y, n2.x FROM close_edges ce '
             'INNER JOIN edges e ON e.id = ce.id2 '
             'INNER JOIN nodes n1 ON n1.id = e.id1 '
             'INNER JOIN nodes n2 ON n2.id = e.id2 '
             'WHERE ce.id1 = ?')
        for l3, l4, p3lat, p3lon, p4lat, p4lon in c.execute(q, (edge_id,)):
            results.append((l3, (p3lat, p3lon), l4, (p4lat, p4lon)))
        return results

    def find_duplicates(self, func=None):
        """Find entries with identical locations."""
        c = self.db.cursor()
        logger.debug('Find duplicates ...')
        t_start = time.time()
        cnt = 0
        q = ('select count(*)as qty, group_concat(id) '
             'from nodes '
             'group by y, x '
             'having qty > 1 ')
        for ncnt, idxs in c.execute(q):
            func(int(idx) for idx in idxs.split(","))
        t_delta = time.time() - t_start
        logger.info(f"Found {cnt} doubles, time: {t_delta} seconds")

    def connect_parallelroads(self, dist=0.5, bb=None):
        c = self.db.cursor()
        it = self.all_edges(bb=bb)
        if tqdm:
            it = tqdm.tqdm(list(it))
        cnt = 0
        for key_a, loc_a, key_b, loc_b in it:
            e_id1 = (key_a, key_b).__hash__()
            bb2 = [min(loc_a[0], loc_b[0]), min(loc_a[1], loc_b[1]),
                   max(loc_a[0], loc_b[0]), max(loc_a[1], loc_b[1])]
            for key_c, loc_c, key_d, loc_d in self.all_edges(bb=bb2):
                e_id2 = (key_c, key_d).__hash__()
                if key_a == key_c or key_a == key_d or key_b == key_c or key_b == key_d:
                    continue
                # print(f"Test: ({key_a},{key_b}) - ({key_c},{key_d})")
                if self.lines_parallel(loc_a, loc_b, loc_c, loc_d, d=dist):
                    # print(f"Parallel: ({key_a},{key_b}) - ({key_c},{key_d})")
                    c.execute('INSERT INTO close_edges(id1, id2) VALUES (?, ?)', (e_id1, e_id2))
                    c.execute('INSERT INTO close_edges(id1, id2) VALUES (?, ?)', (e_id2, e_id1))
                    cnt += 1
        logger.debug(f"Linked {cnt} edges")
        self.db.commit()

    def nodes_to_paths(self, nodes, ignore_nopath=True):
        c = self.db.cursor()
        prev_path = None
        paths = []
        for begin, end in zip(nodes[:-1], nodes[1:]):
            c.execute("SELECT path FROM edges WHERE id1=? AND id2=?", (begin, end))
            path = c.fetchone()[0]
            if path is None and ignore_nopath:
                continue
            if path != prev_path:
                paths.append(path)
                prev_path = path
        return paths

    def path_dist(self, path):
        c = self.db.cursor()
        dist = 0
        q = ('SELECT n1.y, n1.x, n2.y, n2.x FROM edges e '
             'INNER JOIN nodes n1 ON n1.id = e.id1 '
             'INNER JOIN nodes n2 ON n2.id = e.id2 '
             'WHERE e.pathnum>0 AND e.path=?')
        for lat1, lon1, lat2, lon2 in c.execute(q, (path,)):
            dist += self.distance((lat1, lon1), (lat2, lon2))
        return dist

    def print_stats(self):
        print("Graph\n-----")
        print("Nodes: {}".format(len(self.graph)))

    def __str__(self):
        # s = ""
        # for label, (loc, nbrs, _) in self.graph.items():
        #     s += f"{label:<10} - ({loc[0]:10.4f}, {loc[1]:10.4f})\n"
        # return s
        c = self.db.cursor()
        c.execute("select sqlite_version()")
        row = c.fetchone()
        version = row[0]
        return f"SqliteMap({self.name}, size={self.size()}, version={version})"



# encoding: utf-8
"""
leuvenmapmatching.visualization
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:author: Wannes Meert
:copyright: Copyright 2015-2018 DTAI, KU Leuven and Sirris.
:license: Apache License, Version 2.0, see LICENSE for details.
"""
import math
import random
import logging
from itertools import islice

import numpy as np
import matplotlib.pyplot as plt
from matplotlib import colors as mcolors
import smopy


logger = logging.getLogger("be.kuleuven.cs.dtai.mapmatching")
graph_color = mcolors.CSS4_COLORS['darkmagenta']
match_color = mcolors.CSS4_COLORS['green']
match_ne_color = mcolors.CSS4_COLORS['olive']
lattice_color = mcolors.CSS4_COLORS['magenta']
nodes_color = mcolors.CSS4_COLORS['cyan']
path_color = mcolors.CSS4_COLORS['blue']
fontsize = 11  # 7  # 11


def plot_map(map_con, path=None, nodes=None, counts=None, ax=None, use_osm=False, z=None, bb=None,
             show_labels=False, matcher=None, show_graph=False, zoom_path=False, show_lattice=False,
             show_matching=False, filename=None, linewidth=2, coord_trans=None,
             figwidth=20, lattice_nodes=None):
    """Plot the db/graph and optionally include the observed path and inferred nodes.

    :param map_con: Map
    :param path: list[(lat, lon)]
    :param nodes: list[str]
    :param counts: Number of visits of a node in the lattice. dict[label, int]
    :param ax: Matplotlib axis
    :param use_osm: Use OpenStreetMap layer, the points should be latitude-longitude pairs.
    :param matcher: Matcher object (overrules given path, nodes and counts)
    :param filename: File to write image to
    :param show_graph: Plot the vertices and edges in the graph
    :return: None
    """
    if matcher is not None:
        path = matcher.path
        counts = matcher.node_counts()
        nodes = None
        if lattice_nodes is None:
            lat_nodes = matcher.lattice_best
        else:
            lat_nodes = lattice_nodes
        if lat_nodes is None:
            lat_nodes = []
    else:
        lat_nodes = []

    if not bb:
        bb = map_con.bb()
    lat_min, lon_min, lat_max, lon_max = bb
    if path:
        plat, plon = islice(zip(*path), 2)
        lat_min, lat_max = min(lat_min, min(plat)), max(lat_max, max(plat))
        lon_min, lon_max = min(lon_min, min(plon)), max(lon_max, max(plon))
        bb = [lat_min, lon_min, lat_max, lon_max]
    logger.debug("bb = [{}, {}, {}, {}]".format(*bb))

    if zoom_path and path:
        if type(zoom_path) is slice:
            plat, plon = islice(zip(*path[zoom_path]), 2)
            lat_min, lat_max = min(plat), max(plat)
            lon_min, lon_max = min(plon), max(plon)
        else:
            plat, plon = islice(zip(*path), 2)
            lat_min, lat_max = min(plat), max(plat)
            lon_min, lon_max = min(plon), max(plon)
        lat_d = lat_max - lat_min
        lon_d = lon_max - lon_min
        latlon_d = max(lat_d, lon_d)
        lat_max += max(latlon_d * 0.01, lat_d * 0.2)
        lon_min -= max(latlon_d * 0.01, lon_d * 0.2)
        lat_min -= max(latlon_d * 0.01, lat_d * 0.2)
        lon_max += max(latlon_d * 0.01, lon_d * 0.2)
        logger.debug("Setting bounding box to path")
        bb = [lat_min, lon_min, lat_max, lon_max]
        logger.debug("bb(zoom-path) = [{}, {}, {}, {}]".format(*bb))

    bb_o = bb
    if coord_trans:
        logger.debug("Converting bounding box coordinates")
        if path:
            path = [coord_trans(lat, lon) for lat, lon in path]
        lat_min, lon_min, lat_max, lon_max = bb
        lat_min, lon_min = coord_trans(lat_min, lon_min)
        lat_max, lon_max = coord_trans(lat_max, lon_max)
        bb = [lat_min, lon_min, lat_max, lon_max]
        logger.debug("bb = [{}, {}, {}, {}]".format(*bb))

    if use_osm:
        from .util import dist_latlon
        project = dist_latlon.project
        if z is None:
            z = 18
        m = smopy.Map(bb, z=z, ax=ax)
        to_pixels = m.to_pixels
        x_max, y_max = to_pixels(lat_max, lon_max)
        x_min, y_min = to_pixels(lat_min, lon_min)
        height = figwidth / abs(x_max - x_min) * abs(y_max - y_min)
        if ax is None:
            ax = m.show_mpl(figsize=(figwidth, height))
        else:
            ax = m.show_mpl(ax=ax)
        fig = None

    else:
        from .util import dist_euclidean
        project = dist_euclidean.project

        def to_pixels(lat, lon=None):
            if lon is None:
                lat, lon = lat[0], lat[1]
            return lon, lat
        x_max, y_max = to_pixels(lat_max, lon_max)
        x_min, y_min = to_pixels(lat_min, lon_min)
        height = figwidth / abs(lon_max - lon_min) * abs(lat_max - lat_min)
        if ax is None:
            fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(figwidth, height))
        else:
            fig = None
        ax.set_xlim([x_min, x_max])
        ax.set_ylim([y_min, y_max])

    # if counts is None:
    #     node_sizes = [10] * map_con.size()
    # else:
    #     node_sizes = [counts[label]*100+5 for label in map_con.labels()]

    if show_graph:
        logger.debug('Plot vertices ...')
        cnt = 0
        for key, coord in map_con.all_nodes(bb=bb_o):
            if coord_trans:
                coord = coord_trans(*coord)
            coord = to_pixels(coord)
            plt.plot(coord[0], coord[1], marker='o', markersize=2*linewidth, color=graph_color, alpha=0.75)
            if show_labels:
                key = str(key)
                if type(show_labels) is int:
                    key = key[-show_labels:]
                xytext = ax.transLimits.transform(coord)
                xytext = (xytext[0]+0.001, xytext[1]+0.0)
                xytext = ax.transLimits.inverted().transform(xytext)
                # key = str(key)[-3:]
                # print(f'annotate: {key} {coord} {xytext}')
                ann = ax.annotate(key, xy=coord, xytext=xytext,
                            # textcoords=('axes fraction', 'axes fraction'),
                            # arrowprops=dict(arrowstyle='->'),
                            color=graph_color, fontsize=fontsize)
                # ann.set_rotation(45)
            cnt += 1
        logger.debug(f'... done, {cnt} nodes')

        logger.debug('Plot lines ...')
        cnt = 0
        for row in map_con.all_edges(bb=bb_o):
            loc_a = row[1]
            loc_b = row[3]
            if coord_trans:
                loc_a = coord_trans(*loc_a)
                loc_b = coord_trans(*loc_b)
            x_a, y_a = to_pixels(*loc_a)
            x_b, y_b = to_pixels(*loc_b)
            ax.plot([x_a, x_b], [y_a, y_b], color=graph_color, linewidth=linewidth, markersize=linewidth)
            cnt += 1
        logger.debug(f'... done, {cnt} edges')

    if show_lattice:
        if matcher is None:
            logger.warning("Matcher needs to be passed to show lattice. Not showing lattice.")
        plot_lattice(ax, to_pixels, matcher)

    if path:
        logger.debug('Plot path ...')
        if type(zoom_path) is slice:
            path_startidx = zoom_path.start
            path_slice = path[zoom_path]
        else:
            path_startidx = 0
            path_slice = path
        px, py = zip(*[to_pixels(p[:2]) for p in path_slice])
        ax.plot(px, py, linewidth=linewidth, markersize=linewidth * 2, alpha=0.75,
                linestyle="--", marker='o', color=path_color)
        if show_labels:
            for li, (lx, ly) in enumerate(zip(px, py)):
                # ax.text(lx, ly, f"O{li}", color=path_color)
                ann = ax.annotate(f"O{path_startidx + li}", xy=(lx, ly), color=path_color, fontsize=fontsize)
                ann.set_rotation(45)

    if nodes or matcher:
        logger.debug('Plot nodes ...')
        xs, ys, ls = [], [], []
        prev = None
        node_locs = []
        if nodes:
            for node in nodes:
                if type(node) == tuple:
                    node = node[0]
                lat, lon = map_con.node_coordinates(node)
                node_locs.append((lat, lon, node))
        elif lat_nodes is not None:
            prev_m = None
            for m in lat_nodes:
                if prev_m is not None and prev_m.edge_m.l2 == m.edge_m.l1 \
                        and prev_m.edge_m.l1 != m.edge_m.l2:
                    lat, lon = m.edge_m.p1
                    node_locs.append((lat, lon, m.edge_m.l1))
                lat, lon = m.edge_m.pi
                node_locs.append((lat, lon, m.edge_m.label))
                prev_m = m
        for lat, lon, label in node_locs:
            if coord_trans:
                lat, lon = coord_trans(lat, lon)
            if bb[0] <= lat <= bb[2] and bb[1] <= lon <= bb[3]:
                if prev is not None:
                    x, y = to_pixels(*prev)
                    xs.append(x)
                    ys.append(y)
                    ls.append(label)
                    prev = None
                x, y = to_pixels(lat, lon)
                xs.append(x)
                ys.append(y)
                ls.append(label)
            else:
                if prev is None:
                    x, y = to_pixels(lat, lon)
                    xs.append(x)
                    ys.append(y)
                    ls.append(label)
                prev = lat, lon
        ax.plot(xs, ys, 'o-', linewidth=linewidth * 3, markersize=linewidth * 3, alpha=0.75,
                color=nodes_color)
        # if show_labels:
        #     for label, lx, ly in zip(ls, xs, ys):
        #         ax.annotate(label, xy=(lx, ly), xytext=(lx - 30, ly), color=nodes_color)

    if matcher and show_matching:
        logger.debug('Plot matching path-nodes (using matcher) ...')
        for idx, m in enumerate(lat_nodes):
            lat, lon = m.edge_m.pi[:2]
            lat2, lon2 = m.edge_o.pi[:2]
            if coord_trans:
                lat, lon = coord_trans(lat, lon)
                lat2, lon2 = coord_trans(lat2, lon2)
            x, y = to_pixels(lat, lon)
            x2, y2 = to_pixels(lat2, lon2)
            if m.edge_o.is_point():
                plt.plot(x, y, marker='x', markersize=2 * linewidth, color=match_color, alpha=0.75)
                plt.plot(x2, y2, marker='+', markersize=2 * linewidth, color=match_color, alpha=0.75)
                ax.plot((x, x2), (y, y2), '-', color=match_color, linewidth=linewidth, alpha=0.75)
            else:
                plt.plot(x, y, marker='x', markersize=2 * linewidth, color=match_ne_color, alpha=0.75)
                plt.plot(x2, y2, marker='+', markersize=2 * linewidth, color=match_ne_color, alpha=0.75)
                ax.plot((x, x2), (y, y2), '-', color=match_ne_color, linewidth=linewidth, alpha=0.75)
            # ax.plot((x, x2), (y, y2), '-', color=match_color, linewidth=10, alpha=0.1)
            # if show_labels:
            #     ax.annotate(f"{m.obs}.{m.obs_ne}", xy=(x, y))
    elif path and nodes and len(path) == len(nodes) and show_matching:
        logger.debug('Plot matching path-nodes (using sequence of nodes) ...')
        for idx, (loc, node) in enumerate(zip(path, nodes)):
            x, y = to_pixels(*loc)
            if type(node) == tuple and (len(node) == 4 or len(node) == 2):
                latlon2, latlon3 = map_con.node_coordinates(node[0]), map_con.node_coordinates(node[1])
                if coord_trans:
                    latlon2 = coord_trans(*latlon2)
                    latlon3 = coord_trans(*latlon3)
                latlon4, _ = project(latlon2, latlon3, loc)
                x4, y4 = to_pixels(*latlon4)
                ax.plot((x, x4), (y, y4), '-', color=match_color, linewidth=linewidth, alpha=0.75)
            elif type(node) == tuple and len(node) == 3:
                lat2, lon2 = map_con.node_coordinates(node[0])
                if coord_trans:
                    lat2, lon2 = coord_trans(lat2, lon2)
                x2, y2 = to_pixels(lat2, lon2)
                ax.plot((x, x2), (y, y2), '-', color=match_color, linewidth=linewidth, alpha=0.75)
            elif type(node) == str or type(node) == int:
                lat2, lon2 = map_con.node_coordinates(node[0])
                if coord_trans:
                    lat2, lon2 = coord_trans(lat2, lon2)
                x2, y2 = to_pixels(lat2, lon2)
                ax.plot((x, x2), (y, y2), '-', color=match_color, linewidth=linewidth, alpha=0.75)
            else:
                raise Exception('Unknown node type: {} ({})'.format(node, type(node)))
            # if show_labels:
            #     ax.annotate(str(idx), xy=(x, y))
    if map_con.use_latlon:
        ax.set_xlabel('Longitude')
        ax.set_ylabel('Latitude')
    else:
        ax.set_xlabel('X')
        ax.set_ylabel('Y')
    ax.axis('equal')
    ax.set_aspect('equal')
    if filename is not None:
        plt.savefig(filename)
        if fig is not None:
            plt.close(fig)
            fig = None
            ax = None
    return fig, ax


def plot_lattice(ax, to_pixels, matcher):
    for idx in range(len(matcher.lattice)):
        if len(matcher.lattice[idx]) == 0:
            continue
        for m in matcher.lattice[idx].values_all():
            for mp in m.prev:
                if m.stop:
                    alpha = 0.1
                    linewidth = 1
                else:
                    alpha = 0.3
                    linewidth = 3
                if mp.edge_m.p2 is None:
                    prv = mp.edge_m.p1
                else:
                    prv = mp.edge_m.p2
                nxt = m.edge_m.p1
                x1, y1 = to_pixels(*prv)
                x2, y2 = to_pixels(*nxt)
                ax.plot((x1, x2), (y1, y2), '.-', color=lattice_color, linewidth=linewidth, alpha=alpha)
                if m.edge_m.p2 is not None:
                    x1, y1 = to_pixels(*m.edge_m.p1)
                    x2, y2 = to_pixels(*m.edge_m.p2)
                    ax.plot((x1, x2), (y1, y2), '.-', color=lattice_color, linewidth=linewidth, alpha=alpha)


def plot_obs_noise_dist(obs_fn, obs_noise, min_dist=0, max_dist=10):
    """Plot the expected noise of an observation distribution.

    :param matcher: Matcher
    :return:
    """
    x = np.linspace(min_dist, max_dist, 100)
    y = [obs_fn(xi) for xi in x]
    plt.plot(x, y)
    plt.xlabel("Distance")
    plt.ylabel("Probability")
    plt.xlim((min_dist, max_dist))
    plt.ylim((0, 1))
    plt.axvline(x=obs_noise, color='red', alpha=0.7)
    plt.annotate("Observation noise stddev", xy=(obs_noise, 0))



# encoding: utf-8
"""
leuvenmapmatching
~~~~~~~~~~~~~~~~~

:author: Wannes Meert
:copyright: Copyright 2015-2018 DTAI, KU Leuven and Sirris.
:license: Apache License, Version 2.0, see LICENSE for details.
"""
import logging
from . import map, matcher, util
# visualization is not loaded by default (avoid loading unnecessary dependencies such as matplotlib).

logger = logging.getLogger("be.kuleuven.cs.dtai.mapmatching")

__version__ = '1.1.3'




# encoding: utf-8
"""
leuvenmapmatching.util.dist_euclidean
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:author: Wannes Meert
:copyright: Copyright 2015-2018 DTAI, KU Leuven and Sirris.
:license: Apache License, Version 2.0, see LICENSE for details.
"""
import logging
import math

import numpy as np

logger = logging.getLogger("be.kuleuven.cs.dtai.mapmatching")


def distance(p1, p2):
    result = math.sqrt((p1[0] - p2[0]) ** 2 + (p1[1] - p2[1]) ** 2)
    # print("distance({}, {}) -> {}".format(p1, p2, result))
    return result


def distance_point_to_segment(p, s1, s2, delta=0.0):
    p_int, ti = project(s1, s2, p, delta=delta)
    return distance(p_int, p), p_int, ti
    # l1a = np.array(s1)
    # l2a = np.array(s2)
    # pa = np.array(p)
    # return np.linalg.norm(np.cross(l2a - l1a, l1a - pa)) / np.linalg.norm(l2a - l1a)


def distance_segment_to_segment(f1, f2, t1, t2):
    """Distance between segments..

    :param f1: From
    :param f2:
    :param t1: To
    :param t2:
    :return: (distance, proj on f, proj on t, rel pos on f, rel pos on t)
    """
    x1, y1 = f1
    x2, y2 = f2
    x3, y3 = t1
    x4, y4 = t2
    n = ((y4 - y3) * (x2 - x1) - (x4 - x3) * (y2 - y1))
    if np.allclose([n], [0], rtol=0):
        # parallel
        is_parallel = True
        n = 0.0001  # TODO: simulates a point far away
    else:
        is_parallel = False
    u_f = ((x4 - x3) * (y1 - y3) - (y4 - y3) * (x1 - x3)) / n
    u_t = ((x2 - x1) * (y1 - y3) - (y2 - y1) * (x1 - x3)) / n
    xi = x1 + u_f * (x2 - x1)
    yi = y1 + u_f * (y2 - y1)
    changed_f = False
    changed_t = False
    if u_t > 1:
        u_t = 1
        changed_t = True
    elif u_t < 0:
        u_t = 0
        changed_t = True
    if u_f > 1:
        u_f = 1
        changed_f = True
    elif u_f < 0:
        u_f = 0
        changed_f = True
    if not changed_t and not changed_f:
        return 0, (xi, yi), (xi, yi), u_f, u_t
    xf = x1 + u_f * (x2 - x1)
    yf = y1 + u_f * (y2 - y1)
    xt = x3 + u_t * (x4 - x3)
    yt = y3 + u_t * (y4 - y3)
    if changed_t and changed_f:
        # Compare furthest point from intersection with segment
        df = (xf - xi) ** 2 + (yf - yi) ** 2
        dt = (xt - xi) ** 2 + (yt - yi) ** 2
        if df > dt:
            changed_t = False
        else:
            changed_f = False
    if changed_t:
        pt = (xt, yt)
        pf, u_f = project(f1, f2, pt)
    elif changed_f:
        pf = (xf, yf)
        pt, u_t = project(t1, t2, pf)
    else:
        raise Exception(f"Should not happen")
    d = distance(pf, pt)
    return d, pf, pt, u_f, u_t


def project(s1, s2, p, delta=0.0):
    """

    :param s1: Segment start
    :param s2: Segment end
    :param p: Point
    :param delta: Keep delta fraction away from ends
    :return: Point of projection, Relative position on segment
    """
    if np.isclose(s1[0], s2[0], rtol=0) and np.isclose(s1[1], s2[1], rtol=0):
        return s1, 0.0

    l2 = (s1[0]-s2[0])**2 + (s1[1]-s2[1])**2
    t = max(delta, min(1-delta, ((p[0]-s1[0])*(s2[0]-s1[0]) + (p[1]-s1[1])*(s2[1]-s1[1])) / l2))
    return (s1[0] + t * (s2[0]-s1[0]), s1[1] + t * (s2[1]-s1[1])), t


def interpolate_path(path, dd):
    """
    TODO: interplate time as third term
    :param path: (y, x)
    :param dd: Distance difference (meter)
    :return:
    """
    path_new = [path[0]]
    for p1, p2 in zip(path, path[1:]):
        dist = distance(p1, p2)
        if dist > dd:
            dt = int(math.ceil(dist / dd))
            dx = (p2[0] - p1[0]) / dt
            dy = (p2[1] - p1[1]) / dt
            px, py = p1[0], p1[1]
            for _ in range(dt):
                px += dx
                py += dy
                path_new.append((px, py))
        path_new.append(p2)
    return path_new


def box_around_point(p, dist):
    lat, lon = p
    lat_t, lon_r = lat + dist, lon + dist
    lat_b, lon_l = lat - dist, lon - dist
    return lat_b, lon_l, lat_t, lon_r


def lines_parallel(la, lb, lc, ld, d=None):
    x1 = la[0] - lb[0]
    y1 = la[1] - lb[1]
    if x1 == 0:
        if y1 == 0:
            return False
        s1 = 0
    else:
        s1 = math.atan(abs(y1 / x1))
    x2 = lc[0] - ld[0]
    y2 = lc[1] - ld[1]
    if x2 == 0:
        s2 = 0
        if y2 == 0:
            return False
    else:
        s2 = math.atan(abs(y2 / x2))
    thr = math.pi / 180
    if abs(s1 - s2) > thr:
        return False
    if d is not None:
        dist, _, _, _, _ = distance_segment_to_segment(la, lb, lc, ld)
        if dist > d:
            return False
    return True




#!/usr/bin/env python3
# encoding: utf-8
"""
tests.test_path
~~~~~~~~~~~~~~~

:author: Wannes Meert
:copyright: Copyright 2017-2018 DTAI, KU Leuven and Sirris.
:license: Apache License, Version 2.0, see LICENSE for details.
"""
import sys
import os
import logging
from pathlib import Path

import leuvenmapmatching as mm
from leuvenmapmatching.map.inmem import InMemMap
from leuvenmapmatching.matcher.simple import SimpleMatcher
from leuvenmapmatching.matcher.distance import DistanceMatcher


logger = mm.logger
directory = None


def test_path1():
    path = [(0.8, 0.7), (0.9, 0.7), (1.1, 1.0), (1.2, 1.5), (1.2, 1.6), (1.1, 2.0),
            (1.1, 2.3), (1.3, 2.9), (1.2, 3.1), (1.5, 3.2), (1.8, 3.5), (2.0, 3.7),
            (2.1, 3.3), (2.4, 3.2), (2.6, 3.1), (2.9, 3.1), (3.0, 3.2), (3.1, 3.8),
            (3.0, 4.0), (3.1, 4.3), (3.1, 4.6), (3.0, 4.9)]
    # path_sol = ['A', ('A', 'B'), 'B', ('B', 'D'), 'D', ('D', 'E'), 'E', ('E', 'F')]
    path_sol_nodes = ['A', 'B', 'D', 'E', 'F']
    mapdb = InMemMap("map", graph={
        "A": ((1, 1), ["B", "C"]),
        "B": ((1, 3), ["A", "C", "D"]),
        "C": ((2, 2), ["A", "B", "D", "E"]),
        "D": ((2, 4), ["B", "C", "D", "E"]),
        "E": ((3, 3), ["C", "D", "F"]),
        "F": ((3, 5), ["D", "E"])
    }, use_latlon=False)

    matcher = SimpleMatcher(mapdb, max_dist=None, min_prob_norm=None,
                            non_emitting_states=False, only_edges=False)
    path_pred, _ = matcher.match(path, unique=True)
    if directory:
        matcher.print_lattice_stats()
        matcher.print_lattice()
        from leuvenmapmatching import visualization as mmviz
        mmviz.plot_map(mapdb, matcher=matcher, show_labels=True, show_matching=True,
                       show_graph=True, show_lattice=True,
                       filename=str(directory / "test_path1.png"))
    # assert path_pred == path_sol, f"Paths not equal:\n{path_pred}\n{path_sol}"
    nodes_pred = matcher.path_pred_onlynodes
    assert nodes_pred == path_sol_nodes, f"Nodes not equal:\n{nodes_pred}\n{path_sol_nodes}"


def test_path1_dist():
    path = [(0.8, 0.7), (0.9, 0.7), (1.1, 1.0), (1.2, 1.5), (1.2, 1.6), (1.1, 2.0),
            (1.1, 2.3), (1.3, 2.9), (1.2, 3.1), (1.5, 3.2), (1.8, 3.5), (2.0, 3.7),
            (2.1, 3.3), (2.4, 3.2), (2.6, 3.1), (2.9, 3.1), (3.0, 3.2), (3.1, 3.8),
            (3.0, 4.0), (3.1, 4.3), (3.1, 4.6), (3.0, 4.9)]
    # path_sol = ['A', ('A', 'B'), 'B', ('B', 'D'), 'D', ('D', 'E'), 'E', ('E', 'F')]
    path_sol_nodes = ['A', 'B', 'D', 'E', 'F']
    mapdb = InMemMap("map", graph={
        "A": ((1, 1), ["B", "C"]),
        "B": ((1, 3), ["A", "C", "D"]),
        "C": ((2, 2), ["A", "B", "D", "E"]),
        "D": ((2, 4), ["B", "C", "D", "E"]),
        "E": ((3, 3), ["C", "D", "F"]),
        "F": ((3, 5), ["D", "E"])
    }, use_latlon=False)

    matcher = DistanceMatcher(mapdb, max_dist=None, min_prob_norm=None,
                              obs_noise=0.5,
                              non_emitting_states=False)
    matcher.match(path)
    if directory:
        from leuvenmapmatching import visualization as mmviz
        mmviz.plot_map(mapdb, matcher=matcher, show_labels=True, show_matching=True, show_graph=True,
                       filename=str(directory / "test_path1_dist.png"))
    nodes_pred = matcher.path_pred_onlynodes
    assert nodes_pred == path_sol_nodes, f"Nodes not equal:\n{nodes_pred}\n{path_sol_nodes}"


def test_path2():
    path = [(0.8, 0.7), (0.9, 0.7), (1.1, 1.0), (1.2, 1.5), (1.2, 1.6), (1.1, 2.0),
            (1.1, 2.3), (1.3, 2.9), (1.2, 3.1), (1.5, 3.2), (1.8, 3.5), (2.0, 3.7),
            (2.1, 3.3), (2.4, 3.2), (2.6, 3.1), (2.9, 3.1), (3.0, 3.2), (3.1, 3.8),
            (3.0, 4.0), (3.1, 4.3), (3.1, 4.6), (3.0, 4.9)]
    # path_sol = ['A', ('A', 'B'), 'B', ('B', 'D'), 'D', ('D', 'E'), 'E', ('E', 'F')]
    path_sol_nodes = ['A', 'B', 'D', 'E', 'F']
    mapdb = InMemMap("map", graph={
        "A": ((1, 1), ["B", "C", "X"]),
        "B": ((1, 3), ["A", "C", "D", "K"]),
        "C": ((2, 2), ["A", "B", "D", "E", "X", "Y"]),
        "D": ((2, 4), ["B", "C", "F", "E", "K", "L"]),
        "E": ((3, 3), ["C", "D", "F", "Y"]),
        "F": ((3, 5), ["D", "E", "L"]),
        "X": ((2, 0), ["A", "C", "Y"]),
        "Y": ((3, 1), ["X", "C", "E"]),
        "K": ((1, 5), ["B", "D", "L"]),
        "L": ((2, 6), ["K", "D", "F"])
    }, use_latlon=False)

    matcher = SimpleMatcher(mapdb, max_dist=None, min_prob_norm=0.001,
                            non_emitting_states=False, only_edges=False,
                            max_lattice_width=3)
    path_pred, _ = matcher.match(path, unique=True)
    if directory:
        matcher.print_lattice_stats()
        matcher.print_lattice()
        from leuvenmapmatching import visualization as mmviz
        mmviz.plot_map(mapdb, matcher=matcher, show_labels=True, show_matching=True,
                       show_lattice=True, show_graph=True,
                       filename=str(directory / "test_path2.png"))
    # assert path_pred == path_sol, "Nodes not equal:\n{}\n{}".format(path_pred, path_sol)
    nodes_pred = matcher.path_pred_onlynodes
    assert nodes_pred == path_sol_nodes, f"Nodes not equal:\n{nodes_pred}\n{path_sol_nodes}"


def test_path2_inc():
    path = [(0.8, 0.7), (0.9, 0.7), (1.1, 1.0), (1.2, 1.5), (1.2, 1.6), (1.1, 2.0),
            (1.1, 2.3), (1.3, 2.9), (1.2, 3.1), (1.5, 3.2), (1.8, 3.5), (2.0, 3.7),
            (2.1, 3.3), (2.4, 3.2), (2.6, 3.1), (2.9, 3.1), (3.0, 3.2), (3.1, 3.8),
            (3.0, 4.0), (3.1, 4.3), (3.1, 4.6), (3.0, 4.9)]
    # path_sol = ['A', ('A', 'B'), 'B', ('B', 'D'), 'D', ('D', 'E'), 'E', ('E', 'F')]
    path_sol_nodes = ['A', 'B', 'D', 'E', 'F']
    mapdb = InMemMap("map", graph={
        "A": ((1, 1), ["B", "C", "X"]),
        "B": ((1, 3), ["A", "C", "D", "K"]),
        "C": ((2, 2), ["A", "B", "D", "E", "X", "Y"]),
        "D": ((2, 4), ["B", "C", "F", "E", "K", "L"]),
        "E": ((3, 3), ["C", "D", "F", "Y"]),
        "F": ((3, 5), ["D", "E", "L"]),
        "X": ((2, 0), ["A", "C", "Y"]),
        "Y": ((3, 1), ["X", "C", "E"]),
        "K": ((1, 5), ["B", "D", "L"]),
        "L": ((2, 6), ["K", "D", "F"])
    }, use_latlon=False)

    ## Phase 1
    print('=== PHASE 1 ===')
    matcher = SimpleMatcher(mapdb, max_dist=None, min_prob_norm=0.001,
                            non_emitting_states=False, only_edges=False,
                            max_lattice_width=1)
    path_pred, _ = matcher.match(path, unique=True)
    if directory:
        matcher.print_lattice_stats()
        matcher.print_lattice()
        from leuvenmapmatching import visualization as mmviz
        with (directory / 'test_path2_inc_1.gv').open('w') as ofile:
            matcher.lattice_dot(file=ofile, precision=2, render=True)
        mmviz.plot_map(mapdb, matcher=matcher, show_labels=True, show_matching=True,
                       show_lattice=True, show_graph=True,
                       filename=str(directory / "test_path2_inc_1.png"))

    ## Next phases
    for phase_nb, phase_width in enumerate([2, 3]):
        print(f'=== PHASE {phase_nb + 2} ===')
        path_pred, _ = matcher.increase_max_lattice_width(phase_width, unique=True)
        if directory:
            matcher.print_lattice_stats()
            matcher.print_lattice()
            from leuvenmapmatching import visualization as mmviz
            with (directory / f'test_path2_inc_{phase_nb + 2}.gv').open('w') as ofile:
                matcher.lattice_dot(file=ofile, precision=2, render=True)
            mmviz.plot_map(mapdb, matcher=matcher, show_labels=True, show_matching=True,
                           show_lattice=True, show_graph=True,
                           filename=str(directory / f"test_path2_inc_{phase_nb + 2}.png"))

    # assert path_pred == path_sol, "Nodes not equal:\n{}\n{}".format(path_pred, path_sol)
    nodes_pred = matcher.path_pred_onlynodes
    assert nodes_pred == path_sol_nodes, f"Nodes not equal:\n{nodes_pred}\n{path_sol_nodes}"


def test_path2_dist():
    path = [(0.8, 0.7), (0.9, 0.7), (1.1, 1.0), (1.2, 1.5), (1.2, 1.6), (1.1, 2.0),
            (1.1, 2.3), (1.3, 2.9), (1.2, 3.1), (1.5, 3.2), (1.8, 3.5), (2.0, 3.7),
            (2.1, 3.3), (2.4, 3.2), (2.6, 3.1), (2.9, 3.1), (3.0, 3.2), (3.1, 3.8),
            (3.0, 4.0), (3.1, 4.3), (3.1, 4.6), (3.0, 4.9)]
    path_sol_nodes = ['X', 'A', 'B', 'D', 'E', 'F']
    mapdb = InMemMap("map", graph={
        "A": ((1, 1), ["B", "C", "X"]),
        "B": ((1, 3), ["A", "C", "D", "K"]),
        "C": ((2, 2), ["A", "B", "D", "E", "X", "Y"]),
        "D": ((2, 4), ["B", "C", "F", "E", "K", "L"]),
        "E": ((3, 3), ["C", "D", "F", "Y"]),
        "F": ((3, 5), ["D", "E", "L"]),
        "X": ((2, 0), ["A", "C", "Y"]),
        "Y": ((3, 1), ["X", "C", "E"]),
        "K": ((1, 5), ["B", "D", "L"]),
        "L": ((2, 6), ["K", "D", "F"])
    }, use_latlon=False)

    matcher = DistanceMatcher(mapdb, max_dist=None, min_prob_norm=0.001,
                              obs_noise=0.5,
                              non_emitting_states=False)
    matcher.match(path, unique=True)
    if directory:
        from leuvenmapmatching import visualization as mmviz
        mmviz.plot_map(mapdb, matcher=matcher,
                       show_labels=True, show_matching=True, show_graph=True,
                       filename=str(directory / "test_path2_dist.png"))
    nodes_pred = matcher.path_pred_onlynodes
    assert nodes_pred == path_sol_nodes, f"Nodes not equal:\n{nodes_pred}\n{path_sol_nodes}"


def test_path_outlier():
    path = [(0.8, 0.7), (0.9, 0.7), (1.1, 1.0), (1.2, 1.5), (1.2, 1.6), (1.1, 2.0),
            (1.1, 2.3), (1.3, 2.9), (1.2, 3.1), (1.5, 3.2), (1.8, 3.5), (2.0, 3.7),
            (2.1, 3.3), (2.4, 3.2), (2.6, 3.1), (2.9, 3.1), (3.0, 3.2), (3.1, 3.8),
            (3.0, 4.0), (3.1, 4.3), (3.1, 4.6), (3.0, 4.9)]
    path_sol = ['A', 'B', 'D', 'C', 'D', 'E', 'F']
    path.insert(13, (2.3, 1.8))
    mapdb = InMemMap("map", graph={
        "A": ((1, 1), ["B", "C", "X"]),
        "B": ((1, 3), ["A", "C", "D", "K"]),
        "C": ((2, 2), ["A", "B", "D", "E", "X", "Y"]),
        "D": ((2, 4), ["B", "C", "F", "E", "K", "L"]),
        "E": ((3, 3), ["C", "D", "F", "Y"]),
        "F": ((3, 5), ["D", "E", "L"]),
        "X": ((2, 0), ["A", "C", "Y"]),
        "Y": ((3, 1), ["X", "C", "E"]),
        "K": ((1, 5), ["B", "D", "L"]),
        "L": ((2, 6), ["K", "D", "F"])
    }, use_latlon=False)

    matcher = SimpleMatcher(mapdb, max_dist=None, min_prob_norm=0.0001,
                            max_dist_init=1, obs_noise=0.5, obs_noise_ne=10,
                            non_emitting_states=True)
    _, last_idx = matcher.match(path, unique=True)
    path_pred = matcher.path_pred_onlynodes
    if directory:
        matcher.print_lattice_stats()
        matcher.print_lattice()
        from leuvenmapmatching import visualization as mmviz
        with (directory / 'lattice.gv').open('w') as ofile:
            matcher.lattice_dot(file=ofile)
        mmviz.plot_map(mapdb, matcher=matcher, show_labels=True, show_matching=True,
                       filename=str(directory / "test_path_outlier.png"))
        print("Path through lattice:\n" + "\n".join(m.label for m in matcher.lattice_best))
    assert last_idx == len(path) - 1
    assert path_pred == path_sol, "Nodes not equal:\n{}\n{}".format(path_pred, path_sol)


def test_path_outlier2():
    path = [(0.8, 0.7), (0.9, 0.7), (1.1, 1.0), (1.2, 1.5), (1.2, 1.6), (1.1, 2.0),
            (1.1, 2.3), (1.3, 2.9), (1.2, 3.1), (1.5, 3.2), (1.8, 3.5), (2.0, 3.7),
            (2.1, 3.3), (2.4, 3.2), (2.6, 3.1), (2.9, 3.1), (3.0, 3.2), (3.1, 3.8),
            (3.0, 4.0), (3.1, 4.3), (3.1, 4.6), (3.0, 4.9)]
    path.insert(13, (2.3, -3.0))
    mapdb = InMemMap("map", graph={
        "A": ((1, 1), ["B", "C", "X"]),
        "B": ((1, 3), ["A", "C", "D", "K"]),
        "C": ((2, 2), ["A", "B", "D", "E", "X", "Y"]),
        "D": ((2, 4), ["B", "C", "F", "E", "K", "L"]),
        "E": ((3, 3), ["C", "D", "F", "Y"]),
        "F": ((3, 5), ["D", "E", "L"]),
        "X": ((2, 0), ["A", "C", "Y"]),
        "Y": ((3, 1), ["X", "C", "E"]),
        "K": ((1, 5), ["B", "D", "L"]),
        "L": ((2, 6), ["K", "D", "F"])
    }, use_latlon=False)

    matcher = DistanceMatcher(mapdb, max_dist=None, min_prob_norm=0.1,
                            max_dist_init=1, obs_noise=0.25, obs_noise_ne=1,
                            non_emitting_states=True)
    _, last_idx = matcher.match(path, unique=True)
    if directory:
        # matcher.print_lattice_stats()
        # matcher.print_lattice()
        from leuvenmapmatching import visualization as mmviz
        # with (directory / 'lattice.gv').open('w') as ofile:
        #     matcher.lattice_dot(file=ofile)
        mmviz.plot_map(mapdb, matcher=matcher, show_labels=True, show_matching=True,
                       filename=str(directory / "test_path_outlier2.png"))
    assert last_idx == 12


def test_path_outlier_dist():
    path = [(0.8, 0.7), (0.9, 0.7), (1.1, 1.0), (1.2, 1.5), (1.2, 1.6), (1.1, 2.0),
            (1.1, 2.3), (1.3, 2.9), (1.2, 3.1), (1.5, 3.2), (1.8, 3.5), (2.0, 3.7),
            (2.1, 3.3), (2.4, 3.2), (2.6, 3.1), (2.9, 3.1), (3.0, 3.2), (3.1, 3.8),
            (3.0, 4.0), (3.1, 4.3), (3.1, 4.6), (3.0, 4.9)]
    path_sol = ['A', 'B', 'D', 'C', 'E', 'F']
    path.insert(13, (2.3, 1.8))
    mapdb = InMemMap("map", graph={
        "A": ((1, 1), ["B", "C", "X"]),
        "B": ((1, 3), ["A", "C", "D", "K"]),
        "C": ((2, 2), ["A", "B", "D", "E", "X", "Y"]),
        "D": ((2, 4), ["B", "C", "F", "E", "K", "L"]),
        "E": ((3, 3), ["C", "D", "F", "Y"]),
        "F": ((3, 5), ["D", "E", "L"]),
        "X": ((2, 0), ["A", "C", "Y"]),
        "Y": ((3, 1), ["X", "C", "E"]),
        "K": ((1, 5), ["B", "D", "L"]),
        "L": ((2, 6), ["K", "D", "F"])
    }, use_latlon=False)

    matcher = DistanceMatcher(mapdb, max_dist=None, min_prob_norm=0.0001,
                              max_dist_init=1, obs_noise=0.5, obs_noise_ne=10,
                              non_emitting_states=True)
    matcher.match(path)
    path_pred = matcher.path_pred_onlynodes
    if directory:
        from leuvenmapmatching import visualization as mmviz
        mmviz.plot_map(mapdb, matcher=matcher,
                       show_labels=True, show_matching=True, show_graph=True,
                       filename=str(directory / "test_path_outlier_dist.png"))
    # TODO: Smoothing the observation distances could eliminate the outlier
    assert path_pred == path_sol, "Nodes not equal:\n{}\n{}".format(path_pred, path_sol)


def test_path3():
    path = [(3.0, 3.2), (3.1, 3.8), (3.0, 4.0), (3.1, 4.3), (3.1, 4.6), (3.0, 4.9)]
    path_sol = ['E', 'F']
    mapdb = InMemMap("map", graph={
        "E": ((3, 3), ["F"]),
        "F": ((3, 5), ["E"]),
    }, use_latlon=False)

    matcher = SimpleMatcher(mapdb, max_dist=None, min_prob_norm=0.0001,
                            max_dist_init=1, obs_noise=0.25, obs_noise_ne=10,
                            non_emitting_states=True)
    matcher.match(path, unique=True)
    path_pred = matcher.path_pred_onlynodes
    if directory:
        matcher.print_lattice_stats()
        matcher.print_lattice()
        from leuvenmapmatching import visualization as mmviz
        with (directory / 'lattice.gv').open('w') as ofile:
            matcher.lattice_dot(file=ofile)
        mmviz.plot_map(mapdb, matcher=matcher, show_labels=True, show_matching=True,
                       filename=str(directory / "test_path3.png"))
        print("Path through lattice:\n" + "\n".join(m.label for m in matcher.lattice_best))
    assert path_pred == path_sol, "Nodes not equal:\n{}\n{}".format(path_pred, path_sol)


def test_path3_dist():
    path = [(3.0, 3.2), (3.1, 3.8), (3.0, 4.0), (3.1, 4.3), (3.1, 4.6), (3.0, 4.9)]
    path_sol = ['E', 'F']
    mapdb = InMemMap("map", graph={
        "E": ((3, 3), ["F"]),
        "F": ((3, 5), ["E"]),
    }, use_latlon=False)

    matcher = DistanceMatcher(mapdb, max_dist=None, min_prob_norm=0.0001,
                              max_dist_init=1, obs_noise=0.25, obs_noise_ne=10,
                              non_emitting_states=True)
    matcher.match(path, unique=True)
    path_pred = matcher.path_pred_onlynodes
    if directory:
        from leuvenmapmatching import visualization as mmviz
        mmviz.plot_map(mapdb, matcher=matcher, show_labels=True, show_matching=True,
                       filename=str(directory / "test_path3_dist.png"))
        print("Path through lattice:\n" + "\n".join(m.label for m in matcher.lattice_best))
    assert path_pred == path_sol, "Nodes not equal:\n{}\n{}".format(path_pred, path_sol)


def test_path4_dist_inc():
    map_con = InMemMap("mymap", graph={
        "A": ((1, 1), ["B", "C", "X"]),
        "B": ((1, 3), ["A", "C", "D", "K"]),
        "C": ((2, 2), ["A", "B", "D", "E", "X", "Y"]),
        "D": ((2, 4), ["B", "C", "D", "E", "K", "L"]),
        "E": ((3, 3), ["C", "D", "F", "Y"]),
        "F": ((3, 5), ["D", "E", "L"]),
        "X": ((2, 0), ["A", "C", "Y"]),
        "Y": ((3, 1), ["X", "C", "E"]),
        "K": ((1, 5), ["B", "D", "L"]),
        "L": ((2, 6), ["K", "D", "F"])
    }, use_latlon=False)

    path = [(0.8, 0.7), (0.9, 0.7), (1.1, 1.0), (1.2, 1.5), (1.2, 1.6), (1.1, 2.0),
            (1.1, 2.3), (1.3, 2.9), (1.2, 3.1), (1.5, 3.2), (1.8, 3.5), (2.0, 3.7),
            (2.3, 3.5), (2.4, 3.2), (2.6, 3.1), (2.9, 3.1), (3.0, 3.2),
            (3.1, 3.8), (3.0, 4.0), (3.1, 4.3), (3.1, 4.6), (3.0, 4.9)]

    matcher = DistanceMatcher(map_con, max_dist=2, obs_noise=1, min_prob_norm=0.5)
    matcher.match(path[:5])
    if directory:
        from leuvenmapmatching import visualization as mmviz
        mmviz.plot_map(map_con, matcher=matcher,
                       show_labels=True, show_matching=True, show_graph=True,
                       filename=str(directory / "test_path4_dist_inc_1.png"))

    matcher.match(path, expand=True)
    nodes = matcher.path_pred_onlynodes
    if directory:
        from leuvenmapmatching import visualization as mmviz
        mmviz.plot_map(map_con, matcher=matcher,
                       show_labels=True, show_matching=True, show_graph=True,
                       filename=str(directory / "test_path4_dist_inc_2.png"))
    nodes_sol = ['X', 'A', 'B', 'D', 'E', 'F']
    assert nodes == nodes_sol, "Nodes not equal:\n{}\n{}".format(nodes, nodes_sol)


if __name__ == "__main__":
    logger.setLevel(logging.DEBUG)
    logger.addHandler(logging.StreamHandler(sys.stdout))
    directory = Path(os.environ.get('TESTDIR', Path(__file__).parent))
    print(f"Saving files to {directory}")
    # test_path1()
    # test_path1_dist()
    # test_path2()
    # test_path2_inc()
    # test_path2_dist()
    # test_path_outlier()
    # test_path_outlier2()
    # test_path_outlier_dist()
    # test_path3()
    # test_path3_dist()
    test_path4_dist_inc()



#!/usr/bin/env python3
# encoding: utf-8
"""
tests.test_bugs
~~~~~~~~~~~~~~~

:author: Wannes Meert
:copyright: Copyright 2018 DTAI, KU Leuven and Sirris.
:license: Apache License, Version 2.0, see LICENSE for details.
"""
import os
import sys
import logging
from pathlib import Path
import csv

import leuvenmapmatching as mm
from leuvenmapmatching.map.inmem import InMemMap
from leuvenmapmatching.map.sqlite import SqliteMap
from leuvenmapmatching.matcher.simple import SimpleMatcher
from leuvenmapmatching.matcher.distance import DistanceMatcher
import leuvenmapmatching.visualization as mm_viz
MYPY = False
if MYPY:
    from typing import List, Tuple


logger = mm.logger
directory = None


def test_bug1():
    dist = 10
    nb_steps = 20

    map_con = InMemMap("map", graph={
        "A":  ((1, dist), ["B"]),
        "B":  ((2, dist), ["A", "C", "CC"]),
        "C":  ((3, 0), ["B", "D"]),
        "D":  ((4 + dist, 0), ["C", "E"]),
        "CC": ((3, 2 * dist), ["B", "DD"]),
        "DD": ((4 + dist, 2 * dist), ["CC", "E"]),
        "E":  ((5 + dist, dist), ["F", "D", "DD"]),
        "F":  ((6 + dist, dist), ["E", ]),

    }, use_latlon=False)

    i = 10
    path = [(1.1,      2*dist*i/nb_steps),
            (2.1,      2*dist*i/nb_steps),
            (5.1+dist, 2*dist*i/nb_steps),
            (6.1+dist, 2*dist*i/nb_steps)
            # (1, len*i/nb_steps),
            # (2, len*i/nb_steps),
            # (3, len*i/nb_steps)
            ]

    matcher = SimpleMatcher(map_con, max_dist=dist + 1, obs_noise=dist + 1, min_prob_norm=None,
                                  non_emitting_states=True)

    nodes = matcher.match(path, unique=False)
    print("Solution: ", nodes)
    if directory:
        import leuvenmapmatching.visualization as mm_vis
        matcher.print_lattice()
        matcher.print_lattice_stats()
        mm_vis.plot_map(map_con, path=path, nodes=nodes, counts=matcher.node_counts(),
                        show_labels=True, filename=str(directory / "test_bugs_1.png"))


def test_bug2():
    this_path = Path(os.path.realpath(__file__)).parent / "rsrc" / "bug2"
    edges_fn = this_path / "edgesrl.csv"
    nodes_fn = this_path / "nodesrl.csv"
    path_fn = this_path / "path.csv"
    zip_fn = this_path / "leuvenmapmatching_testdata.zip"

    if not (edges_fn.exists() and nodes_fn.exists() and path_fn.exists()):
        import requests
        url = 'https://people.cs.kuleuven.be/wannes.meert/leuvenmapmatching/leuvenmapmatching_testdata.zip'
        logger.debug("Download testfiles from kuleuven.be")
        r = requests.get(url, stream=True)
        with zip_fn.open('wb') as ofile:
            for chunk in r.iter_content(chunk_size=1024):
                if chunk:
                    ofile.write(chunk)
        import zipfile
        logger.debug("Unzipping leuvenmapmatching_testdata.zip")
        with zipfile.ZipFile(str(zip_fn), "r") as zip_ref:
            zip_ref.extractall(str(zip_fn.parent))

    logger.debug(f"Reading map ...")
    mmap = SqliteMap("road_network", use_latlon=True, dir=this_path)

    path = []
    with path_fn.open("r") as path_f:
        reader = csv.reader(path_f, delimiter=',')
        for row in reader:
            lat, lon = [float(coord) for coord in row]
            path.append((lat, lon))
    node_cnt = 0
    with nodes_fn.open("r") as nodes_f:
        reader = csv.reader(nodes_f, delimiter=',')
        for row in reader:
            nid, lonlat, _ = row
            nid = int(nid)
            lon, lat = [float(coord) for coord in lonlat[1:-1].split(",")]
            mmap.add_node(nid, (lat, lon), ignore_doubles=True, no_index=True, no_commit=True)
            node_cnt += 1
    edge_cnt = 0
    with edges_fn.open("r") as edges_f:
        reader = csv.reader(edges_f, delimiter=',')
        for row in reader:
            _eid, nid1, nid2, pid = [int(val) for val in row]
            mmap.add_edge(nid1, nid2, edge_type=0, path=pid, no_index=True, no_commit=True)
            edge_cnt += 1
    logger.debug(f"... done: {node_cnt} nodes and {edge_cnt} edges")
    logger.debug("Indexing ...")
    mmap.reindex_nodes()
    mmap.reindex_edges()
    logger.debug("... done")

    matcher = DistanceMatcher(mmap, min_prob_norm=0.001,
                              max_dist=200, obs_noise=4.07,
                              non_emitting_states=True)
    # path = path[:2]
    nodes, idx = matcher.match(path, unique=True)
    path_pred = matcher.path_pred
    if directory:
        import matplotlib.pyplot as plt
        matcher.print_lattice_stats()
        logger.debug("Plotting post map ...")
        fig = plt.figure(figsize=(100, 100))
        ax = fig.get_axes()
        mm_viz.plot_map(mmap, matcher=matcher, use_osm=True, ax=ax,
                        show_lattice=False, show_labels=True, show_graph=False, zoom_path=True,
                        show_matching=True)
        plt.savefig(str(directory / "test_bug1.png"))
        plt.close(fig)
        logger.debug("... done")


if __name__ == "__main__":
    mm.logger.setLevel(logging.DEBUG)
    mm.logger.addHandler(logging.StreamHandler(sys.stdout))
    directory = Path(os.environ.get('TESTDIR', Path(__file__).parent))
    print(f"Saving files to {directory}")
    # test_bug1()
    test_bug2()



# encoding: utf-8
"""
leuvenmapmatching.map
~~~~~~~~~~~~~~~~~~~~~

:author: Wannes Meert
:copyright: Copyright 2018 DTAI, KU Leuven and Sirris.
:license: Apache License, Version 2.0, see LICENSE for details.
"""



# encoding: utf-8
"""
leuvenmapmatching.map.osmnx
~~~~~~~~~~~~~~~~~~~~~~~~~~~

OSMNX based map.

:author: Wannes Meert
:copyright: Copyright 2022 DTAI, KU Leuven and Sirris.
:license: Apache License, Version 2.0, see LICENSE for details.
"""
import logging
import time
from pathlib import Path
from functools import partial
import osmnx

try:
    import tqdm
except ImportError:
    tqdm = None


from .base import BaseMap


MYPY = False
if MYPY:
    from typing import Optional, Set, Tuple, Dict, Union
    LabelType = Union[int, str]
    LocType = Tuple[float, float]
    EdgeType = Tuple[LabelType, LabelType]


logger = logging.getLogger("be.kuleuven.cs.dtai.mapmatching")


class InMemMap(BaseMap):
    def __init__(self, name, use_latlon=True, use_rtree=False, index_edges=False,
                 crs_lonlat=None, crs_xy=None, graph=None, linked_edges=None, dir=None, deserializing=False):
        """In-memory representation of a map.

        This is a simple database-like object to perform experiments with map matching.
        For production purposes it is recommended to use your own derived
        class (e.g. to connect to your database instance).

        This class supports:

        - Indexing using rtrees to allow for fast searching of points on the map.
          When using the rtree index, only integer numbers are allowed as node labels.
        - Serializing to write and read from files.
        - Projecting points to a different frame (e.g. GPS to Lambert)

        :param name: Map name (mandatory)
        :param use_latlon: The locations represent latitude-longitude pairs, otherwise y-x coordinates
            are assumed.
        :param use_rtree: Build an rtree index to quickly search for locations.
        :param index_edges: Build an index for the edges in the map instead of the vertices.
        :param crs_lonlat: Coordinate reference system for the latitude-longitude coordinates.
        :param crs_xy: Coordiante reference system for the y-x coordinates.
        :param graph: Initial graph of form Dict[label, Tuple[Tuple[y,x], List[neighbor]]]]
        :param dir: Directory where to serialize to. If given, the rtree index structure will be written
            to a file immediately.
        :param deserializing: Internal variable to indicate that the object is being build from a file.
        """
        super(InMemMap, self).__init__(name, use_latlon=use_latlon)
        self.dir = None if dir is None else Path(dir)
        self.index_edges = index_edges
        self.graph = dict() if graph is None else graph
        self.rtree = None
        self.use_rtree = use_rtree
        if self.use_rtree:
            self.setup_index(deserializing=deserializing)

        self.crs_lonlat = 'EPSG:4326' if crs_lonlat is None else crs_lonlat  # GPS
        self.crs_xy = 'EPSG:3395' if crs_xy is None else crs_xy  # Mercator projection
        if pyproj:
            proj_lonlat = pyproj.Proj(self.crs_lonlat, preserve_units=True)
            proj_xy = pyproj.Proj(self.crs_xy, preserve_units=True)
            self.lonlat2xy = partial(pyproj.transform, proj_lonlat, proj_xy)
            self.xy2lonlat = partial(pyproj.transform, proj_xy, proj_lonlat)
        else:
            def pyproj_notfound(*_args, **_kwargs):
                raise Exception("pyproj package not found")
            self.lonlat2xy = pyproj_notfound
            self.xy2lonlat = pyproj_notfound

        self.linked_edges = linked_edges  # type: Optional[Dict[EdgeType, Set[Tuple[EdgeType]]]]
        self.vertex_label_map = None

    def vertex_label_to_int(self, label, create=False):
        if type(label) is int:
            return label
        if self.vertex_label_map is None:
            if not create:
                return label
            self.vertex_label_map = dict()
        if label in self.vertex_label_map:
            new_label = self.vertex_label_map[label]
        else:
            new_label = len(self.vertex_label_map)
            self.vertex_label_map[label] = new_label
        return new_label

    def vertices_labels_to_int(self):
        graph = dict()
        for label, (loc, nbrs) in self.graph.items():
            new_label = self.vertex_label_to_int(label, create=True)
            new_nbrs = [self.vertex_label_to_int(nbr, create=True) for nbr in nbrs]
            graph[new_label] = (loc, new_nbrs)
        self.graph = graph

    def serialize(self):
        """Create a serializable data structure."""
        data = {
            "name": self.name,
            "graph": self.graph,
            "use_latlon": self.use_latlon,
            "use_rtree": self.use_rtree,
            "index_edges": self.index_edges,
            "crs_lonlat": self.crs_lonlat,
            "crs_xy": self.crs_xy,
            "linked_edges": self.linked_edges
        }
        if self.dir is not None:
            data["dir"] = self.dir
        return data

    @classmethod
    def deserialize(cls, data):
        """Create a new instance from a dictionary."""
        nmap = cls(data["name"], dir=data.get("dir", None),
                   use_latlon=data["use_latlon"], use_rtree=data["use_rtree"],
                   index_edges=data["index_edges"],
                   crs_lonlat=data.get("crs_lonlat", None), crs_xy=data.get("crs_xy", None),
                   graph=data.get("graph", None), linked_edges=data.get("linked_edges", None),
                   deserializing=True)
        return nmap

    def dump(self):
        """Serialize map using pickle.

        All files will be saved to the `dir` directory using the `name` as filename.
        """
        if self.dir is None:
            logger.error(f"No directory set where to save (see InMemMap.__init__)")
            return
        filename = self.dir / (self.name + ".pkl")
        with filename.open("wb") as ofile:
            pickle.dump(self.serialize(), ofile)
        logger.debug(f"Saved map to {filename}")
        if self.rtree:
            rtree_fn = self.rtree_fn()
            if rtree_fn is not None:
                self.rtree.close()
                self.rtree = rtree.index.Index(str(rtree_fn))

    @classmethod
    def from_pickle(cls, filename):
        """Deserialize map using pickle to the given filename."""
        filename = Path(filename)
        with filename.open("rb") as ifile:
            data = pickle.load(ifile)
        nmap = cls.deserialize(data)
        return nmap

    def bb(self):
        """Bounding box.

        :return: (lat_min, lon_min, lat_max, lon_max) or (y_min, x_min, y_max, x_max)
        """
        if self.use_rtree:
            lat_min, lon_min, lat_max, lon_max = self.rtree.bounds
        else:
            glat, glon = zip(*[t[0] for t in self.graph.values()])
            lat_min, lat_max = min(glat), max(glat)
            lon_min, lon_max = min(glon), max(glon)
        return lat_min, lon_min, lat_max, lon_max

    def labels(self):
        """All labels."""
        return self.graph.keys()

    def size(self):
        return len(self.graph)

    def node_coordinates(self, node_key):
        """Get the coordinates of the given node.

        :param node_key: Node label/key
        :return: (lat, lon)
        """
        return self.graph[node_key][0]

    def add_node(self, node, loc):
        """Add new node to the map.

        :param node: label
        :param loc: (lat, lon) or (y, x)
        """
        if node in self.graph:
            if self.graph[node][0] is None:
                self.graph[node] = (loc, self.graph[node][1])
        else:
            self.graph[node] = (loc, [])
        if self.use_rtree and self.rtree:
            if type(node) is not int:
                raise Exception(f"Rtree index only supports integer keys for vertices")
            self.rtree.insert(node, (loc[0], loc[1], loc[0], loc[1]))

    def del_node(self, node):
        if node not in self.graph:
            return
        if self.rtree:
            data = self.graph[node]
            loc = data[0]
            self.rtree.delete(node, (loc[0], loc[1], loc[0], loc[1]))
        del self.graph[node]

    def add_edge(self, node_a, node_b):
        """Add new edge to the map.

        :param node_a: Label for the node that is the start of the edge
        :param node_b: Label for the node that is the end of the edge
        """
        if node_a not in self.graph:
            raise ValueError(f"Add {node_a} first as node")
        if node_b not in self.graph:
            raise ValueError(f"Add {node_b} first as node")
        if node_b not in self.graph[node_a][1]:
            self.graph[node_a][1].append(node_b)

    def _items_in_bb(self, bb):
        if self.rtree is not None:
            node_idxs = self.rtree.intersection(bb)
            for key in node_idxs:
                yield (key, self.graph[key])
        else:
            lat_min, lon_min, lat_max, lon_max = bb
            for key, value in self.graph.items():
                ((lat, lon), nbrs) = value
                if lat_min <= lat <= lat_max and lon_min <= lon <= lon_max:
                    yield (key, value)

    def all_edges(self, bb=None):
        """Return all edges.

        :param bb: Bounding box
        :return: (key_a, loc_a, nbr, loc_b)
        """
        if bb is None:
            keyvals = self.graph.items()
        else:
            keyvals = self._items_in_bb(bb)
        for key_a, (loc_a, nbrs) in keyvals:
            if loc_a is not None:
                for nbr in nbrs:
                    try:
                        loc_b, _ = self.graph[nbr]
                        if loc_b is not None:
                            yield (key_a, loc_a, nbr, loc_b)
                    except KeyError:
                        # print("Node not found: {}".format(nbr))
                        pass

    def all_nodes(self, bb=None):
        """Return all nodes.

        :param bb: Bounding box
        :return:
        """
        if bb is None:
            keyvals = self.graph.items()
        else:
            keyvals = self._items_in_bb(bb)
        for key_a, (loc_a, nbrs) in keyvals:
            if loc_a is not None:
                yield key_a, loc_a

    def purge(self):
        cnt_noloc = 0
        cnt_noedges = 0
        remove = []
        for node in self.graph.keys():
            if self.graph[node][0] is None:
                cnt_noloc += 1
                remove.append(node)
                # print("No location for node {}".format(node))
            elif len(self.graph[node][1]) == 0:
                cnt_noedges += 1
                remove.append(node)
        for node in remove:
            self.del_node(node)
        logger.debug("Removed {} nodes without location".format(cnt_noloc))
        logger.debug("Removed {} nodes without edges".format(cnt_noedges))

    def rtree_size(self):
        bb = self.rtree.bounds
        if bb[0] < bb[2] and bb[1] < bb[3]:
            rtree_size = self.rtree.count(bb)
        else:
            rtree_size = 0
        return rtree_size

    def rtree_fn(self):
        rtree_fn = None
        if self.dir is not None:
            rtree_fn = self.dir / self.name
        return rtree_fn

    def setup_index(self, force=False, deserializing=False):
        if not self.use_rtree:
            return
        if self.rtree is not None and not force:
            return
        if rtree is None:
            raise Exception("rtree package not found")

        rtree_fn = self.rtree_fn()
        args = []
        if deserializing and (rtree_fn is None or not rtree_fn.exists()):
            deserializing = False

        if self.graph and len(self.graph) > 0 and not deserializing:
            if self.index_edges:
                logger.debug("Generator to index edges")

                def generator_function():
                    for label, data in self.graph.items():
                        lat_min, lon_min = data[0]
                        lat_max, lon_max = lat_min, lon_min
                        for idx in data[1]:
                            olat, olon = self.graph[idx][0]
                            lat_min = min(lat_min, olat)
                            lat_max = max(lat_max, olat)
                            lon_min = min(lon_min, olon)
                            lon_max = max(lon_max, olon)
                        if type(label) is not int:
                            raise Exception(f"Rtree index only supports integer keys for vertices")
                        yield (label, (lat_min, lon_min, lat_max, lon_max), None)
            else:
                logger.debug("Generator to index nodes")

                def generator_function():
                    for label, data in self.graph.items():
                        lat, lon = data[0]
                        if type(label) is not int:
                            raise Exception(f"Rtree index only supports integer keys for vertices")
                        yield (label, (lat, lon, lat, lon), None)
            args.append(generator_function())

        t_start = time.time()
        if self.dir is not None:
            # props = rtree.index.Property()
            # if force:
            #     props.overwrite = True
            logger.debug(f"Creating new file-based rtree index ({rtree_fn}) ...")
            args.insert(0, str(rtree_fn))
        elif deserializing:
            raise Exception("Cannot deserialize, no directory given")
        else:
            logger.debug(f"Creating new in-memory rtree index (args={args}) ...")
        self.rtree = rtree.index.Index(*args)
        t_delta = time.time() - t_start
        logger.debug(f"... done: rtree size = {self.rtree_size()}, time = {t_delta} sec")

    def fill_index(self):
        if not self.use_rtree or self.rtree is None:
            return

        for label, data in self.graph.items():
            loc = data[0]
            self.rtree.insert(label, (loc[1], loc[0], loc[1], loc[0]))
        logger.debug(f"After filling rtree, size = {self.rtree_size()}")

    def to_xy(self, name=None, use_rtree=None):
        """Create a map that uses a projected XY representation on which Euclidean distances
        can be used.
        """
        if not self.use_latlon:
            return self
        if name is None:
            name = self.name + "_xy"
        if use_rtree is None:
            use_rtree = self.use_rtree

        ngraph = dict()
        for label, row in self.graph.items():
            lat, lon = row[0]
            x, y = self.lonlat2xy(lon, lat)
            ngraph[label] = ((y, x), row[1])
        nmap = self.__class__(name, dir=self.dir, graph=ngraph,
                              use_latlon=False, use_rtree=use_rtree, index_edges=self.index_edges,
                              crs_xy=self.crs_xy, crs_lonlat=self.crs_lonlat)

        return nmap

    def latlon2xy(self, lat, lon):
        x, y = self.lonlat2xy(lon, lat)
        return x, y

    def latlon2yx(self, lat, lon):
        x, y = self.lonlat2xy(lon, lat)
        return y, x

    def xy2latlon(self, x, y):
        lon, lat = self.xy2lonlat(x, y)
        return lat, lon

    def yx2latlon(self, y, x):
        lon, lat = self.xy2lonlat(x, y)
        return lat, lon

    def nodes_closeto(self, loc, max_dist=None, max_elmt=None):
        """Return all nodes close to the given location.

        :param loc: Location
        :param max_dist: Maximal distance from the location
        :param max_elmt: Return only the most nearby nodes
        """
        t_start = time.time()
        lat, lon = loc[:2]
        lat_b, lon_l, lat_t, lon_r = self.box_around_point((lat, lon), max_dist)
        bb = (lat_b, lon_l,  # y_min, x_min
              lat_t, lon_r)  # y_max, x_max
        if self.rtree is not None and max_dist is not None:
            logger.debug(f"Search closeby nodes to {loc}, bb={bb}")
            nodes = self.rtree.intersection(bb)
        else:
            logger.warning("Searching closeby nodes with linear search, use an index and set max_dist")
            if max_dist is not None:
                nodes = (key for key, _ in self._items_in_bb(self.box_around_point((lat, lon), max_dist)))
            else:
                nodes = self.graph.keys()
        t_delta_search = time.time() - t_start
        t_start = time.time()
        results = []
        for label in nodes:
            oloc, nbrs = self.graph[label]
            dist = self.distance(loc, oloc)
            if dist < max_dist:
                results.append((dist, label, oloc))
        results.sort()
        t_delta_dist = time.time() - t_start
        logger.debug(f"Found {len(results)} closeby nodes "
                     f"in {t_delta_search} sec and computed distances in {t_delta_dist} sec")
        if max_elmt is not None:
            results = results[:max_elmt]
        return results

    def edges_closeto(self, loc, max_dist=None, max_elmt=None):
        """Return all nodes that are on an edge that is close to the given location.

        :param loc: Location
        :param max_dist: Maximal distance from the location
        :param max_elmt: Return only the most nearby nodes
        """
        t_start = time.time()
        lat, lon = loc[:2]
        if self.rtree is not None and max_dist is not None and self.index_edges:
            lat_b, lon_l, lat_t, lon_r = self.box_around_point((lat, lon), max_dist)
            bb = (lat_b, lon_l,  # y_min, x_min
                  lat_t, lon_r)  # y_max, x_max
            logger.debug(f"Search closeby edges to {loc}, bb={bb}")
            nodes = self.rtree.intersection(bb)
        else:
            if self.rtree is not None and max_dist is not None and not self.index_edges:
                logger.warning("Index is built for nodes, not for edges, set the index_edges argument to true")
            logger.warning("Searching closeby nodes with linear search, use an index and set max_dist")
            if max_dist is not None:
                bb = self.box_around_point((lat, lon), max_dist)
                nodes = (key for key, _ in self._items_in_bb(bb))
            else:
                nodes = self.graph.keys()
        t_delta_search = time.time() - t_start
        t_start = time.time()
        results = []
        for label in nodes:
            oloc, nbrs = self.graph[label]
            for nbr in nbrs:
                if label == nbr:
                    continue
                nbr_data = self.graph[nbr]
                dist, pi, ti = self.distance_point_to_segment(loc, oloc, nbr_data[0])
                # print(f"label={label}/{oloc}, nbr={nbr}/{nbr_data[0]}   -- loc={loc}  -> {dist}, {pi}, {ti}")
                if dist < max_dist:
                    results.append((dist, label, oloc, nbr, nbr_data[0], pi, ti))
        results.sort()
        t_delta_dist = time.time() - t_start
        logger.debug(f"Found {len(results)} closeby edges "
                     f"in {t_delta_search} sec and computed distances in {t_delta_dist} sec")
        if max_elmt is not None:
            results = results[:max_elmt]
        return results

    def nodes_nbrto(self, node):
        results = []
        if node not in self.graph:
            return results
        loc_node, nbrs = self.graph[node]
        for nbr_label in nbrs + [node]:
            try:
                loc_nbr = self.graph[nbr_label][0]
                if loc_nbr is not None:
                    results.append((nbr_label, loc_nbr))
            except KeyError:
                pass
        return results

    def edges_nbrto(self, edge):
        results = []
        l1, l2 = edge
        p1 = self.node_coordinates(l1)
        p2 = self.node_coordinates(l2)
        # Edges that connect at end of this edge
        for l3, p3 in self.nodes_nbrto(l2):
            results.append((l2, p2, l3, p3))
        # Edges that are in parallel and close
        if self.linked_edges:
            for (l3, l4) in self.linked_edges.get(edge, []):
                p3 = self.node_coordinates(l3)
                p4 = self.node_coordinates(l4)
                results.append((l3, p3, l4, p4))
        return results

    def find_duplicates(self, func=None):
        """Find entries with identical locations."""
        cnt = 0
        for label, data in self.graph.items():
            lat, lon = data[0]
            idxs = list(self.rtree.nearest((lat, lon, lat, lon), num_results=1))
            idxs.remove(label)
            if len(idxs) > 0:
                # logger.info(f"Found doubles for {label}: {idxs}")
                if func:
                    func(label, idxs)
        logger.info(f"Found {cnt} doubles")

    def connect_parallelroads(self, dist=0.5, bb=None):
        if self.rtree is None or not self.index_edges:
            logger.error("Finding parallel roads requires and edge-based index")
            return
        self.linked_edges = {}
        it = self.all_edges(bb=bb)
        if tqdm:
            it = tqdm.tqdm(list(it))
        for key_a, loc_a, key_b, loc_b in it:
            bb2 = [min(loc_a[0], loc_b[0]), min(loc_a[1], loc_b[1]),
                   max(loc_a[0], loc_b[0]), max(loc_a[1], loc_b[1])]
            for key_c, loc_c, key_d, loc_d in self.all_edges(bb=bb2):
                if key_a == key_c or key_a == key_d or key_b == key_c or key_b == key_d:
                    continue
                # print(f"Test: ({key_a},{key_b}) - ({key_c},{key_d})")
                if self.lines_parallel(loc_a, loc_b, loc_c, loc_d, d=dist):
                    # print(f"Parallel: ({key_a},{key_b}) - ({key_c},{key_d})")
                    key = (key_a, key_b)
                    if key in self.linked_edges:
                        self.linked_edges[key].add((key_c, key_d))
                    else:
                        self.linked_edges[key] = {(key_c, key_d)}
        logger.debug(f"Linked {len(self.linked_edges)} edges")

    def print_stats(self):
        print("Graph\n-----")
        print("Nodes: {}".format(len(self.graph)))

    def __str__(self):
        # s = ""
        # for label, (loc, nbrs, _) in self.graph.items():
        #     s += f"{label:<10} - ({loc[0]:10.4f}, {loc[1]:10.4f})\n"
        # return s
        return f"InMemMap({self.name}, size={self.size()})"



#!/usr/bin/env python3
# encoding: utf-8
"""
tests.test_path_onlyedges
~~~~~~~~~~~~~~~~~~~~~~~~~

:author: Wannes Meert
:copyright: Copyright 2018 DTAI, KU Leuven and Sirris.
:license: Apache License, Version 2.0, see LICENSE for details.
"""
import sys
import os
import logging
from pathlib import Path
import leuvenmapmatching as mm
from leuvenmapmatching.map.inmem import InMemMap
from leuvenmapmatching.matcher.simple import SimpleMatcher


logger = mm.logger
directory = None


def test_path1():
    path = [(0.8, 0.7), (0.9, 0.7), (1.1, 1.0), (1.2, 1.5), (1.2, 1.6), (1.1, 2.0),
            (1.1, 2.3), (1.3, 2.9), (1.2, 3.1), (1.5, 3.2), (1.8, 3.5), (2.0, 3.7),
            (2.1, 3.3), (2.4, 3.2), (2.6, 3.1), (2.9, 3.1), (3.0, 3.2), (3.1, 3.8),
            (3.0, 4.0), (3.1, 4.3), (3.1, 4.6), (3.0, 4.9)]
    path_sol = ['A', 'B', 'D', 'E', 'F']
    mapdb = InMemMap("map", graph={
        "A": ((1, 1), ["B", "C"]),
        "B": ((1, 3), ["A", "C", "D"]),
        "C": ((2, 2), ["A", "B", "D", "E"]),
        "D": ((2, 4), ["B", "C", "D", "E"]),
        "E": ((3, 3), ["C", "D", "F"]),
        "F": ((3, 5), ["D", "E"])
    }, use_latlon=False)

    matcher = SimpleMatcher(mapdb, max_dist=None, min_prob_norm=None,
                            only_edges=True, non_emitting_states=False)
    matcher.match(path, unique=True)
    path_pred = matcher.path_pred_onlynodes
    if directory:
        print("Lattice best:")
        for m in matcher.lattice_best:
            print(m)
        matcher.print_lattice_stats()
        matcher.print_lattice()
        from leuvenmapmatching import visualization as mmviz
        mmviz.plot_map(mapdb, matcher=matcher, show_labels=True, show_matching=True,
                       filename=str(directory / "test_onlyedges_path1.png"))
    assert path_pred == path_sol, f"Paths not equal:\n{path_pred}\n{path_sol}"


def test_path3():
    path = [(3.0, 3.2), (3.1, 3.8), (3.0, 4.0), (3.1, 4.3), (3.1, 4.6), (3.0, 4.9)]
    path_sol = ['E', 'F']
    mapdb = InMemMap("map", graph={
        "E": ((3, 3), ["F"]),
        "F": ((3, 5), ["E"]),
    }, use_latlon=False)

    matcher = SimpleMatcher(mapdb, max_dist=None, min_prob_norm=0.0001,
                            max_dist_init=1, obs_noise=0.25, obs_noise_ne=10,
                            non_emitting_states=True,
                            only_edges=True)
    matcher.match(path, unique=True)
    path_pred = matcher.path_pred_onlynodes
    if directory:
        matcher.print_lattice_stats()
        matcher.print_lattice()
        from leuvenmapmatching import visualization as mmviz
        with (directory / 'lattice.gv').open('w') as ofile:
            matcher.lattice_dot(file=ofile)
        mmviz.plot_map(mapdb, matcher=matcher, show_labels=True, show_matching=True,
                       filename=str(directory / "test_onlyedges_path3.png"))
        print("Path through lattice:\n" + "\n".join(m.label for m in matcher.lattice_best))
    assert path_pred == path_sol, "Nodes not equal:\n{}\n{}".format(path_pred, path_sol)


if __name__ == "__main__":
    logger.setLevel(logging.DEBUG)
    logger.addHandler(logging.StreamHandler(sys.stdout))
    directory = Path(os.environ.get('TESTDIR', Path(__file__).parent))
    print(f"Saving files to {directory}")
    test_path1()
    # test_path3()



# encoding: utf-8
"""
leuvenmapmatching.util.openstreetmap
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:author: Wannes Meert
:copyright: Copyright 2015-2018 DTAI, KU Leuven and Sirris.
:license: Apache License, Version 2.0, see LICENSE for details.
"""
import logging
from pathlib import Path
import requests
import tempfile
import osmread

logger = logging.getLogger("be.kuleuven.cs.dtai.mapmatching")


def locations_to_map(locations, map_con, filename=None):
    lats, lons = zip(*locations)
    lon_min, lon_max = min(lons), max(lons)
    lat_min, lat_max = min(lats), max(lats)
    bb = [lon_min, lat_min, lon_max, lat_max]
    return bb_to_map(bb, map_con, filename)


def bb_to_map(bb, map_con, filename=None):
    """Download map from overpass-api.de.

    :param bb: [lon_min, lat_min, lon_max, lat_max]
    :param map:
    :param filename:
    :return:
    """
    if filename:
        xml_file = Path(filename)
    else:
        xml_file = Path(tempfile.gettempdir()) / "osm.xml"
    if not xml_file.exists():
        bb_str = ",".join(str(coord) for coord in bb)
        url = 'http://overpass-api.de/api/map?bbox='+bb_str
        logger.debug("Downloading {} from {} ...".format(xml_file, url))
        r = requests.get(url, stream=True)
        with xml_file.open('wb') as ofile:
            for chunk in r.iter_content(chunk_size=1024):
                if chunk:
                    ofile.write(chunk)
        logger.debug("... done")
    else:
        logger.debug("Reusing existing file: {}".format(xml_file))
    return file_to_map(xml_file, map_con)


def file_to_map(filename, map_con):
    logger.debug("Parse OSM file ...")
    for entity in osmread.parse_file(str(filename)):
        if isinstance(entity, osmread.Way) and 'highway' in entity.tags:
            for node_a, node_b in zip(entity.nodes, entity.nodes[1:]):
                map_con.add_edge(node_a, node_b)
                # Some roads are one-way. We'll add both directions.
                map_con.add_edge(node_b, node_a)
        if isinstance(entity, osmread.Node):
            map_con.add_node(entity.id, (entity.lat, entity.lon))
    logger.debug("... done")
    logger.debug("Purging database ...")
    map_con.purge()
    logger.debug("... done")


def download_map_xml(fn, bbox, force=False, verbose=False):
    """Download map from overpass-api.de based on a given bbox

    :param fn: Filename where to store the map as xml
    :param bbox: String or array with [lon_min, lat_min, lon_max, lat_max]
    :param force: Also download if file already exists
    :param verbose: Verbose output
    :return:
    """
    fn = Path(fn)
    if type(bbox) is list:
        bb_str = ",".join(str(coord) for coord in bbox)
    elif type(bbox) is str:
        bb_str = bbox
    else:
        raise AttributeError('Unknown type for bbox: {}'.format(type(bbox)))
    if force or not fn.exists():
        if verbose:
            print("Downloading {}".format(fn))
        import requests
        url = f'http://overpass-api.de/api/map?bbox={bb_str}'
        r = requests.get(url, stream=True)
        with fn.open('wb') as ofile:
            for chunk in r.iter_content(chunk_size=1024):
                if chunk:
                    ofile.write(chunk)
    else:
        if verbose:
            print("File already exists")


def create_map_from_xml(fn, include_footways=False, include_parking=False):
    """Create an InMemMap from an OpenStreetMap XML file.

    Used for testing routes on OpenStreetMap.
    """
    from ..map.inmem import InMemMap
    map_con = InMemMap("map", use_latlon=True)
    cnt = 0
    ways_filter = ['bridleway', 'bus_guideway', 'track']
    if not include_footways:
        ways_filter += ['footway', 'cycleway', 'path']
    parking_filter = ['driveway']
    if not include_parking:
        parking_filter += ['parking_aisle']
    for entity in osmread.parse_file(str(fn)):
        if isinstance(entity, osmread.Way):
            tags = entity.tags
            if 'highway' in tags \
                and not (tags['highway'] in ways_filter) \
                and not ('access' in tags and tags['access'] == 'private') \
                and not ('landuse' in tags and tags['landuse'] == 'square') \
                and not ('amenity' in tags and tags['amenity'] == 'parking') \
                and not ('service' in tags and tags['service'] in parking_filter) \
                and not ('area' in tags and tags['area'] == 'yes'):
                for node_a, node_b in zip(entity.nodes, entity.nodes[1:]):
                    map_con.add_edge(node_a, node_b)
                    # Some roads are one-way. We'll add both directions.
                    map_con.add_edge(node_b, node_a)
        if isinstance(entity, osmread.Node):
            map_con.add_node(entity.id, (entity.lat, entity.lon))
    map_con.purge()
    return map_con



# encoding: utf-8
"""
leuvenmapmatching.util.projections
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:author: Wannes Meert
:copyright: Copyright 2015-2018 DTAI, KU Leuven and Sirris.
:license: Apache License, Version 2.0, see LICENSE for details.
"""
import math
import logging

import pyproj

logger = logging.getLogger("be.kuleuven.cs.dtai.mapmatching")


def latlon2equirectangular(lat, lon, phi_er=0, lambda_er=0):
    """Naive equirectangular projection.  This is the same as considering (lat,lon) == (y,x).
    This is a lot faster but only works if you are far enough from the poles and the dateline.

    :param lat:
    :param lon:
    :param phi_er: The standard parallels (north and south of the equator) where the scale of the projection is true
    :param lambda_er: The central meridian of the map
    """
    x = (lon - lambda_er) * math.cos(phi_er)
    y = lat - phi_er
    return y, x


def equirectangular2latlon(y, x, phi_er=0, lambda_er=0):
    """Naive equirectangular projection. This is the same as considering (lat,lon) == (y,x).
    This is a lot faster but only works if you are far enough from the poles and the dateline.

    :param phi_er: The standard parallels (north and south of the equator) where the scale of the projection is true
    :param lambda_er: The central meridian of the map
    """
    lon = x / math.cos(phi_er) + lambda_er
    lat = y + phi_er
    return lat, lon


def latlon2grs80(coordinates, lon_0=0.0, lat_ts=0.0, y_0=0, x_0=0.0, zone=31, **kwargs):
    """Given a list of (lon, lat) coordinates, create x-y coordinates in meter.

    :param coordinates: A list of lon-lat tuples
    :param lon_0: Longitude of projection center.
    :param lat_ts: Latitude of true scale. Defines the latitude where scale is not distorted.
    :param y_0: False northing
    :param x_0: False easting
    :param zone: UTM zone to use for projection (Defaults to 31)
    """
    if zone is None:
        # https://en.wikipedia.org/wiki/Universal_Transverse_Mercator_coordinate_system
        zone = 31
    other_options = " ".join(f"+{key}={val}" for key, val in kwargs.items())
    proj = pyproj.Proj(f"+proj=utm +zone={zone} +ellps=GRS80 +units=m "
                       f"+lon_0={lon_0} +lat_ts={lat_ts} +y_0={y_0} +x_0={x_0} "
                       f"+no_defs {other_options}")
    for lon, lat in coordinates:
        x, y = proj(lon, lat)
        yield x, y



# encoding: utf-8
"""
leuvenmapmatching.util.dist_latlon_nvector
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:author: Wannes Meert
:copyright: Copyright 2015-2018 DTAI, KU Leuven and Sirris.
:license: Apache License, Version 2.0, see LICENSE for details.
"""
import logging
import math

import numpy as np
from nvector._core import unit, n_E2lat_lon, great_circle_normal
import nvector as nv


frame = nv.FrameE(a=6371e3, f=0)
logger = logging.getLogger("be.kuleuven.cs.dtai.mapmatching")


def distance(p1, p2):
    """

    :param p1:
    :param p2:
    :return: Distance in meters
    """
    p1 = frame.GeoPoint(p1[0], p1[1], degrees=True)
    p2 = frame.GeoPoint(p2[0], p2[1], degrees=True)
    d, _, _ = p1.distance_and_azimuth(p2)
    # print("distance_latlon({}, {}) -> {}".format(p1, p2, d))
    return d


def distance_gp(p1, p2):
    d, _, _ = p1.distance_and_azimuth(p2)
    return d


def distance_point_to_segment(p, s1, s2, delta=0.0):
    """
    TODO: A point exactly on the line gives an error.

    :param s1: Segment start point
    :param s2: Segment end point
    :param p: Point to measure distance from path to
    :param delta: Stay away from the endpoints with this factor
    :return: (Distance in meters, projected location on segmegnt)
    """
    # TODO: Initialize all points as GeoPoint when loading data
    s1 = frame.GeoPoint(s1[0], s1[1], degrees=True)
    s2 = frame.GeoPoint(s2[0], s2[1], degrees=True)
    p = frame.GeoPoint(p[0], p[1], degrees=True)
    p_int, ti = _project_nvector(s1, s2, p)
    d, _, _ = p.distance_and_azimuth(p_int)
    return d, (p_int.latitude_deg[0], p_int.longitude_deg[0]), ti


def distance_segment_to_segment(f1, f2, t1, t2):
    """Distance between segments. If no intersection within range, simplified to distance from f2 to [t1,t2].

    :param f1: From
    :param f2:
    :param t1: To
    :param t2:
    :return: (distance, proj on f, proj on t, rel pos on t)
    """
    # TODO: Should be improved
    f1_gp = frame.GeoPoint(f1[0], f1[1], degrees=True)
    f2_gp = frame.GeoPoint(f2[0], f2[1], degrees=True)
    path_f = nv.GeoPath(f1_gp, f2_gp)
    t1_gp = frame.GeoPoint(t1[0], t1[1], degrees=True)
    t2_gp = frame.GeoPoint(t2[0], t2[1], degrees=True)
    path_t = nv.GeoPath(t1_gp, t2_gp)
    p_int = path_f.intersect(path_t)
    p_int_gp = p_int.to_geo_point()
    if path_f.on_path(p_int)[0] and path_t.on_path(p_int)[0]:
        # Intersection point is on segments, between both begins and ends
        loc = (p_int_gp.latitude_deg[0], p_int_gp.longitude_deg[0])
        u_f = distance_gp(f1_gp, p_int_gp) / distance_gp(f1_gp, f2_gp)
        u_t = distance_gp(t1_gp, p_int_gp) / distance_gp(t1_gp, t2_gp)
        return 0, loc, loc, u_f, u_t
    # No intersection, use last point of map segment (the assumption is the observations are far apart)
    # TODO: decide which point to use (see distance_segment_to_segment)
    p_int, u_t = _project_nvector(t1_gp, t2_gp, f2_gp)
    u_f = 1
    d, _, _ = f2_gp.distance_and_azimuth(p_int)
    return d, (f1, f2), (p_int_gp.latitude_deg[0], p_int_gp.longitude_deg[0]), u_f, u_t


def project(s1, s2, p, delta=0.0):
    s1 = frame.GeoPoint(s1[0], s1[1], degrees=True)
    s2 = frame.GeoPoint(s2[0], s2[1], degrees=True)
    p = frame.GeoPoint(p[0], p[1], degrees=True)
    p_int, ti = _project_nvector(s1, s2, p, delta=delta)
    return (p_int.latitude_deg[0], p_int.longitude_deg[0]), ti


def _project_nvector(s1, s2, p, delta=0.0):
    path = nv.GeoPath(s1, s2)
    p_intr = _cross_track_point(path, p)
    pin = p_intr.to_nvector().normal
    s1n = s1.to_nvector().normal
    s2n = s2.to_nvector().normal
    ti = np.linalg.norm(pin - s1n) / np.linalg.norm(s2n - s1n)
    ti = max(delta, min(1 - delta, ti))
    return path.interpolate(ti).to_geo_point(), ti


def _cross_track_point(path, point):
    """Extend nvector package to find the projection point.

    The projection point is the closest point on path to the given point.
    Based on the nvector.cross_track_distance function.
    http://www.navlab.net/nvector/

    :param path: GeoPath
    :param point: GeoPoint
    """
    c_E = great_circle_normal(*path.nvector_normals())
    n_EB_E = point.to_nvector().normal  # type: np.array
    c_EP_E = np.cross(c_E, n_EB_E, axis=0)

    # Find intersection point C that is closest to point B
    frame = path.positionA.frame
    n_EA1_E = path.positionA.to_nvector().normal  # should also be ok to use  n_EB_C
    n_EC_E_tmp = unit(np.cross(c_E, c_EP_E, axis=0), norm_zero_vector=np.nan)
    n_EC_E = np.sign(np.dot(n_EC_E_tmp.T, n_EA1_E)) * n_EC_E_tmp
    if np.any(np.isnan(n_EC_E)):
        raise Exception('Paths are Equal. Intersection point undefined. NaN returned.')
    lat_C, long_C = n_E2lat_lon(n_EC_E, frame.R_Ee)
    return nv.GeoPoint(lat_C, long_C, frame=frame)


def interpolate_path(path, dd):
    """
    TODO: interplate time as third term
    :param path: (lat, lon)
    :param dd: Distance difference (meter)
    :return:
    """
    path_new = [path[0]]
    for p1, p2 in zip(path, path[1:]):
        dist = distance(p1, p2)
        if dist > dd:
            s1 = frame.GeoPoint(p1[0], p1[1], degrees=True)
            s2 = frame.GeoPoint(p2[0], p2[1], degrees=True)
            segment = nv.GeoPath(s1, s2)
            dt = int(math.floor(dist / dd))
            for dti in range(1, dt):
                p_new = segment.interpolate(dti/dt).to_geo_point()
                path_new.append((p_new.latitude_deg[0], p_new.longitude_deg[0]))
        path_new.append(p2)
    return path_new
